/* /////////////////////////////////////////////////////////////////////////
 * File:        example_c_2.c
 *
 * Purpose:     C example program for the recls core library. Demonstrates:
 *
 *                - searching (via Recls_SearchFeedback()) for files
 *                - use of multi-part pattern matching
 *                - recursive operation
 *                - display of full path of each entry
 *                - handling of errors and reporting of error information
 *                - elicitation of entry properties via entry structure
 *                  members
 *                - display of progress (of each directory traversed)
 *
 * Created:     29th May 2006
 * Updated:     18th June 2006
 *
 * www:         http://www.recls.org/
 *
 * License:     Copyright (c) 2006, Synesis Software Pty Ltd.
 *              All rights reserved.
 *
 *              (Licensed under the Synesis Software Open License)
 *
 *              This source code is placed into the public domain 2006
 *              by Synesis Software Pty Ltd. There are no restrictions
 *              whatsoever to your use of the software.
 *
 * ////////////////////////////////////////////////////////////////////// */

/* recls Header Files */
#include <recls/recls.h>
#include <recls/alloca.h>

/* STLSoft Header Files */
#ifdef RECLS_PLATFORM_API_WIN32
# include <winstl/system/console_functions.h>
#endif /* RECLS_PLATFORM_API_WIN32 */

/* Standard C Library Files */
#include <stdio.h>      /* for printf() / fprintf()         */
#include <stdlib.h>     /* for EXIT_SUCCESS / EXIT_FAILURE  */
#include <string.h>

/* ////////////////////////////////////////////////////////////////////// */

static void write_backs(FILE *stm, size_t n);
static void write_blanks(FILE *stm, size_t n);
static void write_blank_line(FILE *stm, size_t n);
static size_t get_console_width(void);
static int RECLS_CALLCONV_DEFAULT example_c_2_progress_fn(  recls_char_t const          *dir
                                                        ,   size_t                      dirLen
                                                        ,   recls_process_fn_param_t    param
                                                        ,   void                        *reserved0
                                                        ,   recls_uint32_t              reserved1);

/* ////////////////////////////////////////////////////////////////////// */

struct feedback_t
{
    size_t  lastLen;
};

/* ////////////////////////////////////////////////////////////////////// */

int main()
{
    const char  SEARCH_PATTERN[]    =   "*.h|*.hpp|*.c|*.cpp|*.cs|*.d|*.java|*.pl|*.py|*.rb";

    struct feedback_t   feedback    =   { 0 };
    hrecls_t            hSrch;
    recls_uint32_t      flags       =   RECLS_F_FILES | RECLS_F_RECURSIVE;
    recls_rc_t          rc          =   Recls_SearchFeedback(NULL, SEARCH_PATTERN, flags, example_c_2_progress_fn, &feedback, &hSrch);

    if(RECLS_RC_OK != rc)
    {
        char    err[1001];
        size_t  n   =   Recls_GetErrorString(rc, &err[0], sizeof(err) - 1);

        err[n] = '\0';

        fprintf(stderr, "Search failed: %s\n", err);

        return EXIT_FAILURE;
    }
    else
    {
        recls_info_t    entry;

        Recls_GetDetails(hSrch, &entry);

        do
        {
            write_blank_line(stdout, feedback.lastLen);

            /* full path */
            printf("%s\n", entry->path.begin);

            Recls_CloseDetails(entry);
        }
        while(RECLS_SUCCEEDED(Recls_GetNextDetails(hSrch, &entry)));

        return EXIT_SUCCESS;
    }
}

/* ////////////////////////////////////////////////////////////////////// */

static void write_backs(FILE *stm, size_t n)
{
    char    *backs  =   (char*)recls_alloca(n);

    memset(&backs[0], '\b', n);

    fprintf(stm, "%.*s", n, &backs[0]);
}

static void write_blanks(FILE *stm, size_t n)
{
    char    *blanks =   (char*)recls_alloca(n);

    memset(&blanks[0], ' ', n);

    fprintf(stm, "%.*s", n, &blanks[0]);
}

static void write_blank_line(FILE *stm, size_t n)
{
    char    *backs  =   (char*)recls_alloca(n);
    char    *blanks =   (char*)recls_alloca(n);

    memset(&backs[0], '\b', n);
    memset(&blanks[0], ' ', n);

    fprintf(stm, "%.*s", n, &backs[0]);
    fprintf(stm, "%.*s", n, &blanks[0]);
    fprintf(stm, "%.*s", n, &backs[0]);
}

static size_t get_console_width(void)
{
    /* In reality, this should evaluate as appropriate to the operating
     * system.
     */

#ifdef RECLS_PLATFORM_API_WIN32
    return winstl__get_console_width();
#else /* ? RECLS_PLATFORM_API_??? */
    return 48;
#endif /* RECLS_PLATFORM_API_??? */
}

static int RECLS_CALLCONV_DEFAULT example_c_2_progress_fn(  recls_char_t const          *dir
                                                        ,   size_t                      dirLen
                                                        ,   recls_process_fn_param_t    param
                                                        ,   void                        *reserved0
                                                        ,   recls_uint32_t              reserved1)
{
    struct feedback_t   *feedback       =   (struct feedback_t*)param;
    size_t              newLen;
    size_t              cch;
    size_t              consoleWidth    =   get_console_width() - 1;

    if(consoleWidth < dirLen)
    {
        char    *squeezedForm = (char*)recls_alloca(dirLen);

        strncpy(squeezedForm, dir, dirLen);

        squeezedForm[dirLen - 1] = '\0';

        cch = Recls_SqueezePath(squeezedForm, squeezedForm, consoleWidth - 1);

        dir = squeezedForm;
    }
    else
    {
        cch = dirLen;
    }

    write_backs(stdout, feedback->lastLen);

    newLen = fprintf(stdout, "%.*s", (int)cch, dir);

    if(newLen < feedback->lastLen)
    {
        size_t  spare   =   feedback->lastLen - newLen;

        write_blanks(stdout, spare);
        write_backs(stdout, spare);
    }

    feedback->lastLen = newLen;

    ((void)reserved0);
    ((void)reserved1);

    return 1; /* Continue processing. */
}

/* ////////////////////////////////////////////////////////////////////// */
