/* /////////////////////////////////////////////////////////////////////////////
 * File:        example_c_3.c
 *
 * Purpose:     C example program for the recls core library. Demonstrates:
 *
 *                - stat() of current directory (via Recls_Stat())
 *                - searching (via Recls_Search()) for files and directories
 *                - recursive operation
 *                - display of full path of each entry, squeezed to constant
 *                  width (via Recls_SqueezePath())
 *                - display of file size for file entries
 *                - display of directory contents size for directory
 *                  entries, (determined via Recls_CalcDirectoryEntrySize())
 *                - handling of errors and reporting of error information
 *                - elicitation of entry properties via API function calls
 *
 * Created:     29th May 2006
 * Updated:     18th June 2006
 *
 * www:         http://www.recls.org/
 *
 * License:     Copyright (c) 2006, Synesis Software Pty Ltd.
 *              All rights reserved.
 *
 *              (Licensed under the Synesis Software Open License)
 *
 *              This source code is placed into the public domain 2006
 *              by Synesis Software Pty Ltd. There are no restrictions
 *              whatsoever to your use of the software.
 *
 * ////////////////////////////////////////////////////////////////////// */

/* recls Header Files */
#include <recls/recls.h>

/* Standard C Library Files */
#include <stdio.h>      /* for printf() / fprintf()         */
#include <stdlib.h>     /* for EXIT_SUCCESS / EXIT_FAILURE  */
#include <string.h>

/* /////////////////////////////////////////////////////////////////////////
 * Macros and definitions
 */

#ifndef RECLS_NUM_ELEMENTS
# define RECLS_NUM_ELEMENTS(x)          (sizeof(x) / sizeof((x)[0]))
#endif /* !RECLS_NUM_ELEMENTS */

/* ////////////////////////////////////////////////////////////////////// */

int main()
{
    /* stat() the current directory */
    recls_info_t    current;
    recls_rc_t      rc  =   Recls_Stat(".", RECLS_F_DIRECTORIES | RECLS_F_DIRECTORY_PARTS, &current);

    if(RECLS_FAILED(rc))
    {
        /* The search failed. Display the error string. */
        char    err[1001];
        size_t  n   =   Recls_GetErrorString(rc, &err[0], sizeof(err) - 1);

        err[n] = '\0';

        fprintf(stderr, "stat of current directory failed: %s\n", err);

        return EXIT_FAILURE;
    }
    else
    {
        hrecls_t        hSrch;
        recls_uint32_t  flags   =   RECLS_F_FILES | RECLS_F_DIRECTORIES | RECLS_F_RECURSIVE;

        rc = Recls_Search(current->path.begin, NULL, flags, &hSrch);

        /* ... close the entry handle, ... */
        Recls_CloseDetails(current);

        if(RECLS_RC_NO_MORE_DATA == rc)
        {
            printf("  no matches found\n");

            return EXIT_SUCCESS;
        }
        else if(RECLS_FAILED(rc))
        {
            /* The search failed. Display the error string. */
            char    err[1001];
            size_t  n   =   Recls_GetErrorString(rc, &err[0], sizeof(err) - 1);

            err[n] = '\0';

            fprintf(stderr, "Search failed: %s\n", err);

            return EXIT_FAILURE;
        }
        else
        {
            /* Get the details for the first entry, ... */ 

            recls_info_t    entry;

            Recls_GetDetails(hSrch, &entry);

            do
            {
                /* ... get the full path, ... */
                recls_filesize_t    size;
                int                 isDirectory;
                unsigned long       ulSize;
                char                path[1001];
                char                squeezedPath[48];
                size_t              cch = Recls_GetPathProperty(entry, &path[0], RECLS_NUM_ELEMENTS(path) - 1);;

                path[cch] = '\0';

                /* ... squeeze it into 48 characters, ... */
                cch = Recls_SqueezePath(path, &squeezedPath[0], RECLS_NUM_ELEMENTS(squeezedPath));

                /* ... determine type, ... */
                isDirectory = Recls_IsFileDirectory(entry);

                if(isDirectory)
                {
                    /* ... calculate size, or ... */
                    size = Recls_CalcDirectoryEntrySize(entry);
                }
                else
                {
                    /* ... elicit size, ... */

                    Recls_GetSizeProperty(entry, &size);
                }

#if defined(RECLS_PLATFORM_IS_WIN32)
                ulSize = (unsigned long)size.QuadPart;
#else /* ? OS */
                ulSize = (unsigned long)size;
#endif /* OS */

                printf( "%.*s: %s; %lu bytes\n"
                    ,   (int)cch
                    ,   squeezedPath
                    ,   isDirectory ? "directory" : "file"
                    ,   ulSize);

                /* ... close the entry handle, ... */
                Recls_CloseDetails(entry);

            } /* ... and get the next entry. */
            while(RECLS_SUCCEEDED(Recls_GetNextDetails(hSrch, &entry)));

            /* Close the search handle. */
            Recls_SearchClose(hSrch);

            return EXIT_SUCCESS;
        }
    }
}

/* ////////////////////////////////////////////////////////////////////// */
