/* /////////////////////////////////////////////////////////////////////////////
 * File:        example_stl_2.cpp
 *
 * Purpose:     STL example program for recls/STL. Demonstrates:
 *
 *                - searching for files
 *                - recursive operation
 *                - display of full path of each entry
 *                - handling exceptions and reporting of error information
 *                - elicitation of entry properties via method calls
 *
 * Created:     18th June 2006
 * Updated:     18th June 2006
 *
 * www:         http://www.recls.org/
 *
 * License:     Copyright (c) 2006, Synesis Software Pty Ltd.
 *              All rights reserved.
 *
 *              (Licensed under the Synesis Software Open License)
 *
 *              This source code is placed into the public domain 2006
 *              by Synesis Software Pty Ltd. There are no restrictions
 *              whatsoever to your use of the software.
 *
 * ////////////////////////////////////////////////////////////////////// */


/* recls Header Files */
#include <recls/stl/root_sequence.hpp>
#include <recls/implicit_link.h>
#include <recls/alloca.h>

/* STLSoft Header Files */
#ifdef RECLS_PLATFORM_API_WIN32
# include <winstl/system/console_functions.h>
#endif /* RECLS_PLATFORM_API_WIN32 */

/* Standard C++ Header Files */
#include <exception>
#include <iostream>
#include <iterator>

/* Standard C Header Files */
#include <stdlib.h>     /* for EXIT_SUCCESS / EXIT_FAILURE  */

/* ////////////////////////////////////////////////////////////////////// */

static void write_backs(FILE *stm, size_t n);
static void write_blanks(FILE *stm, size_t n);
static void write_blank_line(FILE *stm, size_t n);
static size_t get_console_width(void);
static int RECLS_CALLCONV_DEFAULT example_c_2_progress_fn(  recls::recls_char_t const       *dir
                                                        ,   size_t                          dirLen
                                                        ,   recls::recls_process_fn_param_t param
                                                        ,   void                            *reserved0
                                                        ,   recls::recls_uint32_t           reserved1);

/* ////////////////////////////////////////////////////////////////////// */

struct feedback_t
{
    size_t  lastLen;
};

/* ////////////////////////////////////////////////////////////////////////// */

int main()
{
    try
    {
        feedback_t                  feedback    =   { 0 };
        unsigned                    rootTypes   =   recls::RECLS_ROOTS_F_FIXED_DRIVES
                                                |   recls::RECLS_ROOTS_F_CDROM_DRIVES
                                                |   recls::RECLS_ROOTS_F_RAM_DRIVES;
        recls::stl::root_sequence   roots(rootTypes);

        std::cout << "Selected roots:" << std::endl;
        std::copy(  roots.begin(), roots.end()
                ,   std::ostream_iterator<recls::stl::root_sequence::value_type>(std::cout, "\n"));

        std::cout << std::endl;
        std::cout << "Determining size of selected roots ..." << std::endl;

        { for(recls::stl::root_sequence::const_iterator b = roots.begin(); b != roots.end(); ++b)
        {
            recls::recls_filesize_t size = recls::Recls_CalcDirectorySizeFeedback(*b, recls::IGNORE_HIDDEN_ENTRIES_ON_WIN32, example_c_2_progress_fn, &feedback);

            write_blank_line(stdout, feedback.lastLen);

            std::cout   << *b
                        << " => "
#ifdef RECLS_PLATFORM_API_WIN32
                        << static_cast<unsigned long>(size.QuadPart)
#else /* ? RECLS_PLATFORM_API_??? */
                        << static_cast<unsigned long>(size)
#endif /* RECLS_PLATFORM_API_??? */
                         << " bytes"
                        << std::endl;
        }}
    }
    catch(recls::stl::recls_exception &x)
    {
        std::cerr << "Recls error: " << x.rc() << ", " << x.what() << std::endl;

        return EXIT_FAILURE;
    }
    catch(std::bad_alloc &)
    {
        std::cerr << "Out of memory" << std::endl;

        return EXIT_FAILURE;
    }
    catch(std::exception &x)
    {
        std::cerr << "Unhandled error: " << x.what() << std::endl;

        return EXIT_FAILURE;
    }
    catch(...)
    {
        std::cerr << "Unhandled unknown error" << std::endl;

        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

/* ////////////////////////////////////////////////////////////////////////// */

static void write_backs(FILE *stm, size_t n)
{
    char    *backs  =   (char*)recls_alloca(n);

    memset(&backs[0], '\b', n);

    fprintf(stm, "%.*s", n, &backs[0]);
}

static void write_blanks(FILE *stm, size_t n)
{
    char    *blanks =   (char*)recls_alloca(n);

    memset(&blanks[0], ' ', n);

    fprintf(stm, "%.*s", n, &blanks[0]);
}

static void write_blank_line(FILE *stm, size_t n)
{
    char    *backs  =   (char*)recls_alloca(n);
    char    *blanks =   (char*)recls_alloca(n);

    memset(&backs[0], '\b', n);
    memset(&blanks[0], ' ', n);

    fprintf(stm, "%.*s", n, &backs[0]);
    fprintf(stm, "%.*s", n, &blanks[0]);
    fprintf(stm, "%.*s", n, &backs[0]);
}

static size_t get_console_width(void)
{
    /* In reality, this should evaluate as appropriate to the operating
     * system.
     */

#ifdef RECLS_PLATFORM_API_WIN32
    return winstl::get_console_width();
#else /* ? RECLS_PLATFORM_API_??? */
    return 48;
#endif /* RECLS_PLATFORM_API_??? */
}

static int RECLS_CALLCONV_DEFAULT example_c_2_progress_fn(  recls::recls_char_t const       *dir
                                                        ,   size_t                          dirLen
                                                        ,   recls::recls_process_fn_param_t param
                                                        ,   void                            *reserved0
                                                        ,   recls::recls_uint32_t           reserved1)
{
    struct feedback_t   *feedback       =   (struct feedback_t*)param;
    size_t              newLen;
    size_t              cch;
    size_t              consoleWidth    =   get_console_width() - 1;

    if(consoleWidth < dirLen)
    {
        char    *squeezedForm = (char*)recls_alloca(dirLen);

        strncpy(squeezedForm, dir, dirLen);

        squeezedForm[dirLen - 1] = '\0';

        cch = recls::Recls_SqueezePath(squeezedForm, squeezedForm, consoleWidth - 1);

        dir = squeezedForm;
    }
    else
    {
        cch = dirLen;
    }

    write_backs(stdout, feedback->lastLen);

    newLen = fprintf(stdout, "%.*s", (int)cch, dir);

    if(newLen < feedback->lastLen)
    {
        size_t  spare   =   feedback->lastLen - newLen;

        write_blanks(stdout, spare);
        write_backs(stdout, spare);
    }

    feedback->lastLen = newLen;

    ((void)reserved0);
    ((void)reserved1);

    return 1; /* Continue processing. */
}

/* ////////////////////////////////////////////////////////////////////// */
