/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls/cpp/filesearch.hpp
 *
 * Purpose:     recls C++ mapping - FileSearch class.
 *
 * Created:     18th August 2003
 * Updated:     13th August 2006
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2006, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef RECLS_INCL_RECLS_CPP_HPP_FILESEARCH
#define RECLS_INCL_RECLS_CPP_HPP_FILESEARCH

/* File version */
#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_RECLS_CPP_HPP_FILESEARCH_MAJOR       3
# define RECLS_VER_RECLS_CPP_HPP_FILESEARCH_MINOR       2
# define RECLS_VER_RECLS_CPP_HPP_FILESEARCH_REVISION    1
# define RECLS_VER_RECLS_CPP_HPP_FILESEARCH_EDIT        24
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/** \file recls/cpp/filesearch.hpp
 *
 * \brief [C++] Definition of the
 *  \link recls::cpp::FileSearch FileSearch\endlink
 *  class, part of the
 *  \ref group__recls__cpp "recls C++ mapping"
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls/cpp/recls.hpp>
#include <recls/cpp/search.hpp>
#include <recls/assert.h>

#include <stlsoft/memory/auto_buffer.hpp>
#include <stlsoft/shims/access/string.hpp>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{

namespace cpp
{
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// This class provides functionality for executing a file search
///
/// \ingroup group__recls__cpp
class FileSearch
    : public Search
{
/// \name Types
public:
/// @{
    typedef recls_char_t    char_type;
    typedef size_t          size_type;
    typedef FileSearch      class_type;
/// @}

/// \name Construction
/// @{
public:
    /// Creates a search for the items identified by \c pattern in the directory \c rootDir, according to the given \c flags.
    ///
    /// \param rootDir The starting directory for the search. If NULL, or the empty string, uses the current directory
    /// \param pattern The search pattern, e.g. "*.h". If NULL, or the empty string, uses "*.*" (Win32) or "*" (UNIX)
    /// \param flags A combination of enumerants from the \c RECLS_FLAG enumeration
    FileSearch(char_type const *rootDir, char_type const *pattern, recls_uint32_t flags);
    /// Destructor
    ~FileSearch();
/// @}

/// \name Operations
/// @{
public:
    /// Retrieves the information for the given path
    ///
    /// \param path Absolute or relative path of the file system entry whose information is to be retrieved.
    /// \param flags Flags to moderate the search. Only the type flags (flags & RECLS_F_TYPEMASK) and RECLS_F_DIRECTORY_PARTS are recognised
    ///
    /// \return A valid instance if the path/flags exists, or a 'null' instance. (A null instance \c e evaluates \c !e to be true).
    template <typename S>
    static FileEntry    Stat(S const &path)
    {
        return Stat_(stlsoft::c_str_ptr(path), RECLS_F_FILES | RECLS_F_DIRECTORIES | RECLS_F_DIRECTORY_PARTS);
    }

    template <typename S>
    static FileEntry    Stat(   S const         &path
                            ,   recls_uint32_t  flags /* = RECLS_F_FILES | RECLS_F_DIRECTORIES | RECLS_F_DIRECTORY_PARTS */)
    {
        return Stat_(stlsoft::c_str_ptr(path), flags);
    }

    template<   typename    S1
            ,   typename    S2
            >
    static string_t     CombinePaths(   S1 const    &path1
                                    ,   S2 const    &path2)
    {
        return CombinePaths_(stlsoft::c_str_ptr(path1), stlsoft::c_str_ptr(path2));
    }

    template<   typename    S1
            ,   typename    S2
            >
    static string_t     DeriveRelativePath( S1 const    &origin
                                        ,   S2 const    &target)
    {
        return DeriveRelativePath_(stlsoft::c_str_ptr(origin), stlsoft::c_str_ptr(target));
    }

    template <typename S>
    static string_t     SqueezePath(    S const     &path
                                    ,   size_type   width)
    {
        return SqueezePath_(stlsoft::c_str_ptr(path), width);
    }
/// @}

// Implementation
private:
    static ctor_args_type CreateSearch_(char_type const *rootDir, char_type const *pattern, recls_uint32_t flags)
    {
        hrecls_t    hSrch;
        recls_rc_t  rc = Recls_Search(rootDir, pattern, flags, &hSrch);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        if( RECLS_FAILED(rc) &&
            RECLS_RC_NO_MORE_DATA != rc)
        {
            throw ReclsException(rc);
        }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return Search::ctor_args_type(hSrch, rc);
    }

    static FileEntry Stat_(char_type const *path, recls_uint32_t flags);

    static string_t CombinePaths_( char_type const *path1, char_type const *path2);

    static string_t DeriveRelativePath_(char_type const *origin, char_type const *target);

    static string_t SqueezePath_(char_type const *path, size_type width);

// Not to be implemented
private:
    FileSearch(class_type const &rhs);
    class_type &operator =(class_type const &rhs);
};

/* /////////////////////////////////////////////////////////////////////////////
 * Implementation
 */

inline /* static */ string_t FileSearch::CombinePaths_(FileSearch::char_type const *path1, FileSearch::char_type const *path2)
{
    stlsoft::auto_buffer<char_type> buffer(Recls_CombinePaths(path1, path2, NULL, 0));

    Recls_CombinePaths(path1, path2, &buffer[0], buffer.size());

    return string_t(buffer.data(), buffer.size());
}

inline /* static */ string_t FileSearch::DeriveRelativePath_(FileSearch::char_type const *origin, FileSearch::char_type const *target)
{
    stlsoft::auto_buffer<char_type> buffer(Recls_DeriveRelativePath(origin, target, NULL, 0));

    Recls_DeriveRelativePath(origin, target, &buffer[0], buffer.size());

    return string_t(buffer.data(), buffer.size());
}

inline /* static */ string_t FileSearch::SqueezePath_(FileSearch::char_type const *path, size_type width)
{
    stlsoft::auto_buffer<char_type> buffer(width);
    size_t                          cch =   Recls_SqueezePath(path, &buffer[0], buffer.size());

    RECLS_ASSERT(cch <= buffer.size());

    return string_t(buffer.data(), cch);
}


inline FileSearch::FileSearch(FileSearch::char_type const *rootDir, FileSearch::char_type const *pattern, recls_uint32_t flags)
    : Search(CreateSearch_(rootDir, pattern, flags))
{}

inline FileSearch::~FileSearch()
{}

inline /* static */ FileEntry FileSearch::Stat_(FileSearch::char_type const *path, recls_uint32_t flags /* = RECLS_F_FILES | RECLS_F_DIRECTORIES | RECLS_F_DIRECTORY_PARTS */)
{
    recls_info_t    info;
    recls_rc_t      rc  =   Recls_Stat(path, flags, &info);

    return RECLS_SUCCEEDED(rc) ? make_entry_(info) : FileEntry();
}

#if 0
inline /* static */ string_t FileSearch::CombinePaths(FileSearch::char_type const *path1, FileSearch::char_type const *path2)
{
    return CombinePaths_(path1, path2);
}

inline /* static */ string_t FileSearch::DeriveRelativePath(FileSearch::char_type const *origin, FileSearch::char_type const *target)
{
    return DeriveRelativePath_(origin, target);
}

inline /* static */ string_t FileSearch::SqueezePath(FileSearch::char_type const *path, size_type width)
{
    return SqueezePath_(path, width);
}
#endif /* 0 */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace cpp */
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !RECLS_INCL_RECLS_CPP_HPP_FILESEARCH */

/* ////////////////////////////////////////////////////////////////////////// */
