/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls/internal/platform_types.h
 *
 * Purpose:     Platform discrimination for the recls API.
 *
 * Created:     18th August 2003
 * Updated:     16th June 2006
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2006, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef RECLS_INCL_RECLS_INTERNAL_H_PLATFORM_TYPES
#define RECLS_INCL_RECLS_INTERNAL_H_PLATFORM_TYPES

/* File version */
#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_RECLS_INTERNAL_H_PLATFORM_TYPES_MAJOR     3
# define RECLS_VER_RECLS_INTERNAL_H_PLATFORM_TYPES_MINOR     1
# define RECLS_VER_RECLS_INTERNAL_H_PLATFORM_TYPES_REVISION  1
# define RECLS_VER_RECLS_INTERNAL_H_PLATFORM_TYPES_EDIT      22
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/** \file recls/internal/platform_types.h
 *
 * \brief [C, C++] Platform-dependent types for the
 *  \ref group__recls API.
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef RECLS_INCL_RECLS_INTERNAL_H_PLATFORM
# error recls/internal/platform_types.h must not be included directly. You should include recls/recls.h
#endif /* !RECLS_INCL_RECLS_INTERNAL_H_PLATFORM */

#if defined(RECLS_PURE_API) || \
    defined(RECLS_PLATFORM_IS_UNIX)
# include <time.h>
# include <sys/types.h>
#elif defined(RECLS_PLATFORM_IS_WIN32)
# include <windows.h>
#elif defined(RECLS_OVERRIDE_PLATFORM)
 /* You're expected to ensure that the appropriate inclusions are made */
#else
# error Platform not (yet) recognised
#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Platform-dependent types
 */

/** \def recls_time_t
 *
 * \brief The time type for the recls API.
 */
/** \def recls_filesize_t
 *
 * \brief The file-size type for the recls API.
 */

#if defined(RECLS_PURE_API) || \
    defined(RECLS_PLATFORM_IS_UNIX)

 typedef time_t                     recls_time_t;
 typedef off_t                      recls_filesize_t;

#elif defined(RECLS_PLATFORM_IS_WIN32)

 typedef FILETIME                   recls_time_t;
 typedef ULARGE_INTEGER             recls_filesize_t;

#elif defined(RECLS_OVERRIDE_PLATFORM)

# ifndef RECLS_CUSTOM_TIME_T
#  error Must define RECLS_CUSTOM_TIME_T if compiling with RECLS_OVERRIDE_PLATFORM
# endif /* !RECLS_CUSTOM_TIME_T */

# ifndef RECLS_CUSTOM_FILESIZE_T
#  error Must define RECLS_CUSTOM_FILESIZE_T if compiling with RECLS_OVERRIDE_PLATFORM
# endif /* !RECLS_CUSTOM_TIME_T */

 typedef RECLS_CUSTOM_TIME_T        recls_time_t;
 typedef RECLS_CUSTOM_FILESIZE_T    recls_filesize_t;

#else

# error Platform not (yet) recognised

 typedef <platform-dependent-type>  recls_time_t;
 typedef <platform-dependent-type>  recls_filesize_t;

#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

#if defined(_UNICODE) && \
    !defined(UNICODE)
# define UNICODE
#endif /* _UNICODE && !UNICODE */

#if defined(UNICODE) && \
    !defined(_UNICODE)
# define _UNICODE
#endif /* UNICODE && !_UNICODE */

#ifdef RECLS_CHAR_TYPE_IS_CHAR
# undef RECLS_CHAR_TYPE_IS_CHAR
#endif /* RECLS_CHAR_TYPE_IS_CHAR */

#ifdef RECLS_CHAR_TYPE_IS_WCHAR
# undef RECLS_CHAR_TYPE_IS_WCHAR
#endif /* RECLS_CHAR_TYPE_IS_WCHAR */

/** \brief The recls library ambient character type */
#if defined(UNICODE) && \
    !defined(RECLS_NO_UNICODE)
typedef recls_char_w_t              recls_char_t;
# define RECLS_CHAR_TYPE_IS_WCHAR
# define RECLS_LITERAL(x)           L ## x
#else /* ? UNICODE && !RECLS_NO_UNICODE */
typedef recls_char_a_t              recls_char_t;
# define RECLS_CHAR_TYPE_IS_CHAR
# define RECLS_LITERAL(x)           x
#endif /* UNICODE && !RECLS_NO_UNICODE */


/** \brief An asymmetric range representing a sequence of characters (ie a string) */
struct recls_strptrs_t
{
    /** Points to the start of the sequence. */
    recls_char_t const  *begin;
    /** Points to one-past-the-end of the sequence. */
    recls_char_t const  *end;
};

/** \brief An asymmetric range representing a sequence of recls_strptrs_t (ie a collection of strings) */
struct recls_strptrsptrs_t
{
    /** Points to the start of the sequence. */
    struct recls_strptrs_t const    *begin;
    /** Points to one-past-the-end of the sequence. */
    struct recls_strptrs_t const    *end;
};

/** A file-system entry info structure
 *
 * \note Several parts of this structure are platform-dependent.
 */
#if !defined(RECLS_PURE_API)
struct recls_fileinfo_t
{
/** \name attributes */
/** @{ */
    /** \brief The entry's attributes */
    recls_uint32_t              attributes;
/** @} */
/** \name Path components  */
/** @{ */
    /** \brief The full path of the entry.
     *
     * \note This is always nul-terminated.
     */
    struct recls_strptrs_t      path;
# if defined(RECLS_PLATFORM_IS_WIN32)
    /** \brief The short (8.3) path of the entry
     *
     * \note This is always nul-terminated.
     * \note This member is only defined for the Win32 platform.
     */
    struct recls_strptrs_t      shortFile;
    /** \brief The letter of the drive */
    recls_char_t                drive;
# endif /* RECLS_PLATFORM_IS_WIN32 */
    /** \brief The directory component.
     *
     * \note This is <b>never<b> nul-terminated.
     */
    struct recls_strptrs_t      directory;
    /** \brief The file name component (excluding extension).
     *
     * \note This is <b>never<b> nul-terminated.
     */
    struct recls_strptrs_t      fileName;
    /** \brief The file extension component (excluding '.').
     *
     * \note You should <b>never<b> assume that this entry is nul-terminated.
     */
    struct recls_strptrs_t      fileExt;
    /** \brief The directory parts.
     *
     * \note This parts are <b>never<b> nul-terminated.
     */
    struct recls_strptrsptrs_t  directoryParts;
/** @} */
/** \name File times  */
/** @{ */
# if defined(RECLS_PLATFORM_IS_WIN32)
    /** \brief The time the entry was created
     *
     * \note This member is only defined for the Win32 platform.
     */
    recls_time_t                creationTime;
# endif /* RECLS_PLATFORM_IS_WIN32 */
    /** \brief The time the entry was last modified */
    recls_time_t                modificationTime;
    /** \brief The time the entry was last accessed */
    recls_time_t                lastAccessTime;
# if defined(RECLS_PLATFORM_IS_UNIX)
    /** \brief The time the entry status was last changed
     *
     * \note This member is only defined for the UNIX platform.
     */
    recls_time_t                lastStatusChangeTime;
# endif /* RECLS_PLATFORM_IS_UNIX */
/** @} */
/** \name Size  */
/** @{ */
    /** \brief The size of the entry */
    recls_filesize_t            size;
/** @} */
/** \name Path components (2) */
/** @{ */
    /** \brief The search directory specified in the search.
     *
     * \note This is <b>never<b> nul-terminated.
     */
    struct recls_strptrs_t      searchDirectory;
    /** \brief The opaque data of the entry; it is not accessible to any 
     * client code, and <b>must not be manipulated</b> in any way
     */
    struct recls_strptrs_t      searchRelativePath;
/** @} */
#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
/* data */
    /** \brief Reserved for future use. */
    recls_uint64_t              checkSum;
    /** \brief Reserved for future use. */
    recls_uint32_t              extendedFlags[2];
    /** \brief The full path of the entry, relative to the directory
     * from which it was searched.
     *
     * \note This is always nul-terminated.
     */
    recls_byte_t                data[1];
    /*
     *
     * - full path
     * - directory parts
     *
     */
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */
};
#endif /* !RECLS_PURE_API */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !RECLS_INCL_RECLS_INTERNAL_H_PLATFORM_TYPES */

/* ////////////////////////////////////////////////////////////////////////// */
