////////////////////////////////////////////////////////////////////////////////
// File:	Search.cs
//
// Created: 3rd June 2004
// Updated: 15th March 2005
//
////////////////////////////////////////////////////////////////////////////////

namespace recls
{
	using System;
	using System.Collections;

	/// <summary>
	/// Represents a file search operation
	/// </summary>
	/// <remarks>Each Search instance captures the search directory, pattern and flags of a particular 
	/// search. The Search instance may then be used in two ways:
	/// <list type="number">
	/// <item><description>as the expression of a <c>foreach</c> statement</description></item>
	/// <item><description>as the generator of an <see cref="IEnumerator"/>, via the <see cref="GetEnumerator"/> method, which 
	/// may then be used explicitly to provide enumeration of the search results via the <c>MoveNext</c> and <c>Reset</c>
	/// methods, and the <c>Current</c> property.
	/// </description></item>
	/// </list>
	/// </remarks>
	public abstract class Search
		: IDisposable
	{
		/// <summary>
		/// The delegate for file entry processing
		/// </summary>
		/// <param name="entry">The current entry</param>
		/// <returns>Return true to continue processing, or false to cancel all further processing</returns>
		/// <remarks>This delegate is called for each entry in a search initiated by the <see cref="FileSearch.Process"/> and <see cref="FtpSearch.Process"/> methods.</remarks>
		public delegate bool ProcessEntry(FileEntry entry);

		/// <summary>
		/// Returns the enumerator with which the given search can be enumerated
		/// </summary>
		/// <returns>An IEnumerator for the entire Search.</returns>
		/// <remarks>This method enables a Search instance to be used in a 
		/// foreach expression, as in:
		/// <code>
		/// 
		/// Search search = . . .;
		/// foreach(<see cref="FileEntry"/> entry in search)
		/// {
		///   Console.WriteLine(entry.<see cref="FileEntry.Path"/>);
		/// }
		/// 
		/// </code>
		/// As with the <see cref="Search"/> constructor, no enumeration is carried out at this time, only
		/// when the returned enumerator is used.
		/// 
		/// When the enumerator's <c>MoveNext</c> method is called, it may throw a <see cref="ReclsException"/> if the
		/// search fails, except for the case where the search runs out of data. In this case, the enumerator's 
		/// <c>MoveNext</c> method returns false.
		/// </remarks>
		public abstract IEnumerator GetEnumerator();

		/// <summary>
		/// Frees the resources associated with the search
		/// </summary>
		public abstract void Dispose();
	}
}
