/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls.d
 *
 * Purpose:     D mapping for the recls library. recls is a platform-independent
 *              recursive search library. It is mapped to several languages, 
 *              including D. recls was written by Matthew Wilson, as the first
 *              exemplar for his "Positive Integration" column in C/C++ User's
 *              Journal.
 *
 * Created      10th Octover 2003
 * Updated:     16th June 2006
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/** \file recls.d The "recls D mapping" file, containing the definitions of the 
 * \link recls.FileSearch FileSearch\endlink, 
 * \link recls.FtpSearch FtpSearch\endlink,
 * and 
 * \link recls.Entry Entry\endlink,
 * classes.
 */

////////////////////////////////////////////////////////////////////////////////
// Module

module recls;

////////////////////////////////////////////////////////////////////////////////
// Imports

private import std.date;
private import std.string;
version(linux)
{
    private import std.c.linux.linux;
}
else version(Windows)
{
    private import std.c.windows.windows;
// Phobos: Current version does not have WindowsException
//    private import std.windows.exceptions;
}
else
{
    const int platform_not_discriminated = 0;

    static assert(platform_not_discriminated);
}

////////////////////////////////////////////////////////////////////////////////
// Version

const uint  RECLS_VER_MAJOR     =   1;
const uint  RECLS_VER_MINOR     =   8;
const uint  RECLS_VER_REVISION  =   1;
const uint  RECLS_VER_1_0_1     =   0x01000100;
const uint  RECLS_VER_1_1_1     =   0x01010100;
const uint  RECLS_VER_1_2_1     =   0x01020100;
const uint  RECLS_VER_1_3_1     =   0x01030100;
const uint  RECLS_VER_1_4_1     =   0x01040100;
const uint  RECLS_VER_1_4_2     =   0x01040200;
const uint  RECLS_VER_1_5_1     =   0x01050100;
const uint  RECLS_VER_1_5_2     =   0x01050200;
const uint  RECLS_VER_1_5_3     =   0x01050300;
const uint  RECLS_VER_1_6_1     =   0x01060100;
const uint  RECLS_VER_1_6_2     =   0x01060200;
const uint  RECLS_VER_1_6_3     =   0x01060300;
const uint  RECLS_VER_1_6_4     =   0x01060400;
const uint  RECLS_VER_1_6_5     =   0x01060500;
const uint  RECLS_VER_1_7_1     =   0x01070100;
const uint  RECLS_VER_1_8_1     =   0x01080100;
const uint  RECLS_VER           =   RECLS_VER_1_8_1;

////////////////////////////////////////////////////////////////////////////////
// Public types

private alias int               recls_sint32_t;
/// Unsigned 32-bit integer, used for flags
public alias uint               recls_uint32_t;
/// boolean type, used in the recls API functions
public alias int                recls_bool_t;
/// boolean type, used in the recls class mappings
public alias recls_bool_t       boolean;


version(linux)
{
    /// UNIX time type
    public alias time_t         recls_time_t;

    /// UNIX file size type
    public alias off_t          recls_filesize_t;
}
else version(Windows)
{
    /// Win32 time type
    public alias FILETIME       recls_time_t;

    /// Win32 file size type
    public alias ulong          recls_filesize_t;
}
else
{
    const int platform_not_discriminated = 0;

    static assert(platform_not_discriminated);
}

/// The recls search handle type.
public typedef void             *hrecls_t;
/// The recls entry handle type.
public typedef void             *recls_info_t;
/// The recls entry process callback function parameter type. */    
public typedef void             *recls_process_fn_param_t;

/// The return code of the recls API
public typedef recls_sint32_t   recls_rc_t;

/// Returns non-zero if the given return code represents a failure condition.
public boolean RECLS_FAILED(recls_rc_t rc)
{
    return rc < 0;
}

/// Returns non-zero if the given return code represents a success condition.
public boolean RECLS_SUCCEEDED(recls_rc_t rc)
{
    return !RECLS_FAILED(rc);
}

////////////////////////////////////////////////////////////////////////////////
// Private types

private const uint  RECLS_ROOT_NAME_LEN     =   3;

private struct recls_root_t
{
    char    name[1 + RECLS_ROOT_NAME_LEN];
};

////////////////////////////////////////////////////////////////////////////////
// Values

/** General success code */
public const recls_rc_t         RECLS_RC_OK             =   cast(recls_rc_t)(0);

/** Return code that indicates that there is no more data available from an otherwise valid search. */
public const recls_rc_t         RECLS_RC_NO_MORE_DATA   =   cast(recls_rc_t)(-1004);

/// The flags used to moderate the recls search behaviour
public enum RECLS_FLAG
{
    /* These are the recommended forms */
        FILES                       =   0x00000001 /*!< Include files in search. Included by default if none specified */
    ,   DIRECTORIES                 =   0x00000002 /*!< Include directories in search. Not currently supported. */
    ,   LINKS                       =   0x00000004 /*!< Include links in search. Ignored in Win32. */
    ,   DEVICES                     =   0x00000008 /*!< Include devices in search. Not currently supported. */
    ,   TYPEMASK                    =   0x00000FFF
    ,   RECURSIVE                   =   0x00010000 /*!< Searches given directory and all sub-directories */
    ,   NO_FOLLOW_LINKS             =   0x00020000 /*!< Does not expand links */
    ,   DIRECTORY_PARTS             =   0x00040000 /*!< Fills out the directory parts. Supported from version 1.1.1 onwards. */
    ,   DETAILS_LATER               =   0x00080000 /*!< Does not fill out anything other than the path. Not currently supported. */
    ,   PASSIVE_FTP                 =   0x00100000 /*!< Passive mode in FTP */
    ,   MARK_DIRS                   =   0x00200000 /*!< Marks the directories with a trailing slash. */
    ,   ALLOW_REPARSE_DIRS          =   0x00400000 /*!< Allow Win32 reparse point directories to be examined (which can cause infinite loops). */
    ,   CALLBACKS_STDCALL_ON_WIN32  =   0x01000000 /*!< Treats the callbacks as stdcall calling convention on Win32. Default is cdecl. */
	,	USE_TILDE_ON_NO_SEARCHROOT  =   0x04000000 /*!< Interprets a NULL or empty searchRoot as the home directory, rather than the current directory. */

    /* These will be phased out in a subsequent release, leaving only the none-prefixed ones */
    ,   RECLS_F_FILES               =   FILES           /*!< Obsolete. Instead use FILES           */
    ,   RECLS_F_DIRECTORIES         =   DIRECTORIES     /*!< Obsolete. Instead use DIRECTORIES     */
    ,   RECLS_F_LINKS               =   LINKS           /*!< Obsolete. Instead use LINKS           */
    ,   RECLS_F_DEVICES             =   DEVICES         /*!< Obsolete. Instead use DEVICES         */
    ,   RECLS_F_TYPEMASK            =   TYPEMASK        /*!< Obsolete. Instead use TYPEMASK        */
    ,   RECLS_F_RECURSIVE           =   RECURSIVE       /*!< Obsolete. Instead use RECURSIVE       */
    ,   RECLS_F_NO_FOLLOW_LINKS     =   NO_FOLLOW_LINKS /*!< Obsolete. Instead use NO_FOLLOW_LINKS */
    ,   RECLS_F_DIRECTORY_PARTS     =   DIRECTORY_PARTS /*!< Obsolete. Instead use DIRECTORY_PARTS */
    ,   RECLS_F_DETAILS_LATER       =   DETAILS_LATER   /*!< Obsolete. Instead use DETAILS_LATER   */
    ,   RECLS_F_PASSIVE_FTP         =   PASSIVE_FTP     /*!< Obsolete. Instead use PASSIVE_FTP     */
};

/// Flags used to moderate the behaviour of the getSelectedRoots() method
public enum RECLS_ROOTS_FLAG
{
        FIXED_DRIVES				=   0x0001  /*!< Include local fixed drives. */
    ,   NETWORK_DRIVES				=   0x0002  /*!< Include network drives. */
    ,   CDROM_DRIVES				=   0x0004  /*!< Include CD-ROM / DVD-ROM drives. */
    ,   REMOVABLE_DRIVES			=   0x0008  /*!< Include removable drives. */
    ,   RAM_DRIVES					=   0x0010  /*!< Include RAM drives. */
};

////////////////////////////////////////////////////////////////////////////////
// Private recls API declarations

extern (C)
{
    private recls_rc_t Recls_Search(    char            *searchRoot
                                    ,   char            *pattern
                                    ,   recls_uint32_t  flags
                                    ,   hrecls_t        *phSrch);

    private recls_rc_t Recls_SearchFtp( char            *host
                                    ,   char            *username
                                    ,   char            *password
                                    ,   char            *searchRoot
                                    ,   char            *pattern
                                    ,   recls_uint32_t  flags
                                    ,   hrecls_t        *phSrch);

    typedef int (*hrecls_process_fn_t)(recls_info_t info, recls_process_fn_param_t param);

    private recls_rc_t Recls_SearchProcess( char                        *searchRoot
                                        ,   char                        *pattern
                                        ,   recls_uint32_t  flags
                                        ,   hrecls_process_fn_t         pfn
                                        ,   recls_process_fn_param_t    param);


    typedef int (*hrecls_progress_fn_t)(char                        *dir
                                    ,   size_t                      dirLen
                                    ,   recls_process_fn_param_t    param
                                    ,   void                        *reserved0
                                    ,   recls_uint32_t              reserved1);

    private recls_rc_t  Recls_SearchFeedback(   char                        *searchRoot
                                    ,           char                        *pattern
                                    ,           recls_uint32_t              flags
                                    ,           hrecls_progress_fn_t        pfn
                                    ,           recls_process_fn_param_t    param
                                    ,           hrecls_t                    *phSrch);

    private void Recls_SearchClose(in hrecls_t hSrch);

    private recls_rc_t  Recls_GetNext(in hrecls_t hSrch);
    private recls_rc_t  Recls_GetDetails(in hrecls_t hSrch, out recls_info_t pinfo);
    private recls_rc_t  Recls_GetNextDetails(in hrecls_t hSrch, out recls_info_t pinfo);

    private void        Recls_CloseDetails(in recls_info_t fileInfo);
    private recls_rc_t  Recls_CopyDetails(in recls_info_t fileInfo, in recls_info_t *pinfo);
    private recls_rc_t  Recls_OutstandingDetails(in hrecls_t hSrch, out recls_uint32_t count);

    private recls_rc_t  Recls_GetLastError(in hrecls_t hSrch);
    private int         Recls_GetErrorString(in recls_rc_t rc, in char *buffer, in uint cchBuffer);
    private int         Recls_GetLastErrorString(in hrecls_t hSrch, in char *buffer, in uint cchBuffer);

    private uint            Recls_GetPathProperty(in recls_info_t  fileInfo, in char *buffer, in uint cchBuffer);
    private uint            Recls_GetSearchRelativePathProperty(in recls_info_t  fileInfo, in char *buffer, in uint cchBuffer);
version(Windows)
{
    private void            Recls_GetDriveProperty(in recls_info_t fileInfo, out char chDrive);
} // version Windows
    private uint            Recls_GetDirectoryProperty(in recls_info_t fileInfo, in char *buffer, in uint cchBuffer);
    private uint            Recls_GetDirectoryPathProperty(in recls_info_t fileInfo, in char *buffer, in uint cchBuffer);
	private uint            Recls_GetSearchDirectoryProperty(in recls_info_t fileInfo, in char *buffer, in uint cchBuffer);
    private uint            Recls_GetUNCDriveProperty(in recls_info_t fileInfo, in char *buffer, in uint cchBuffer);
    private uint            Recls_GetFileProperty(in recls_info_t fileInfo, in char *buffer, in uint cchBuffer);
    private uint            Recls_GetShortFileProperty(in recls_info_t fileInfo, in char *buffer, in uint cchBuffer);
    private uint            Recls_GetFileNameProperty(in recls_info_t fileInfo, in char *buffer, in uint cchBuffer);
    private uint            Recls_GetFileExtProperty(in recls_info_t fileInfo, in char *buffer, in uint cchBuffer);
    private uint            Recls_GetDirectoryPartProperty(in recls_info_t fileInfo, in int part, in char *buffer, in uint cchBuffer);
    private boolean         Recls_IsFileReadOnly(in recls_info_t fileInfo);
    private boolean         Recls_IsFileDirectory(in recls_info_t fileInfo);
    private boolean         Recls_IsFileLink(in recls_info_t fileInfo);
    private boolean         Recls_IsFileUNC(in recls_info_t fileInfo);
    private void            Recls_GetSizeProperty(in recls_info_t fileInfo, recls_filesize_t *size);
    private recls_time_t    Recls_GetCreationTime(in recls_info_t fileInfo);
    private recls_time_t    Recls_GetModificationTime(in recls_info_t fileInfo);
    private recls_time_t    Recls_GetLastAccessTime(in recls_info_t fileInfo);
    private recls_time_t    Recls_GetLastStatusChangeTime(in recls_info_t fileInfo);

    private uint            Recls_GetRoots(recls_root_t *roots, uint cRoots);
    private uint            Recls_GetSelectedRoots(recls_root_t *roots, uint cRoots, uint flags);
    char                    *Recls_GetPathNameSeparator();
    char                    *Recls_GetPathSeparator();
    char                    *Recls_GetWildcardsAll();

    private recls_rc_t      Recls_Stat(in char *buffer, in uint flags, in recls_info_t *pinfo);

    private uint            Recls_CombinePaths(in char *path1, in char *path2, in char *result, in uint cchResult);
    private uint            Recls_DeriveRelativePath(in char *origin, in char *target, in char *result, in uint cchResult);
    private uint            Recls_SqueezePath(in char *path, in char *result, in uint cchResult);
}

////////////////////////////////////////////////////////////////////////////////
// Variables

version(Windows)
{

//int useWfuncs = 1;
int useWfuncs = 0;  // Not currently supported

/+ ++ Not currently supported ++
static this()
{
    // Win 95, 98, ME do not implement the W functions
    useWfuncs = (GetVersion() < 0x80000000);
}
+/

private char[] MBStoUTF8(char[] s)
{
    wchar[] wbuf;
    int     n;

    if(null is s)
    {
        return null;
    }

    wbuf.length =   MultiByteToWideChar(0, 0, &s[0], s.length, null, 0);
    n           =   MultiByteToWideChar(0, 0, &s[0], wbuf.length, cast(wchar*)wbuf, wbuf.length);

    assert(n == wbuf.length);

    return std.utf.toUTF8(wbuf);
}

private char[] MBSztoUTF8(char *s)
{
    if(null !is s)
    {
        return MBStoUTF8(s[0 .. std.string.strlen(s)]);
    }
    else
    {
        return null;
    }
}

} // version(Windows)

////////////////////////////////////////////////////////////////////////////////
// Helper functions

private d_time recls_time2d_time(recls_time_t rt)
{
version(linux)
{
    return cast(d_time)(rt);
}
else version(Windows)
{
    SYSTEMTIME  st;

    if(!FileTimeToSystemTime(&rt, &st))
    {
// Phobos: Current version does not have WindowsException
//        throw new WindowsException("Invalid time format");
    }

    return SYSTEMTIME2d_time(&st, 0);
}
else
{
    const int platform_not_discriminated = 0;

    static assert(platform_not_discriminated);
} // version
}

private boolean isRoot_(char[] dir)
{
version(linux)
{
    if("/" == dir)
    {
        return true;
    }
}
else version(Windows)
{
    if( 3 <= dir.length &&
        (   '\\' == dir[2] ||
            '/'  == dir[2]) &&
        ':' == dir[1] &&
        std.string.isalpha(dir[0]) &&
        (   dir.length == 3 ||
            '\0' == dir[3]))
    {
        return true;
    }
}
else
{
    const int platform_not_discriminated = 0;

    static assert(platform_not_discriminated);
} // version

    return false;
}

////////////////////////////////////////////////////////////////////////////////
// Public functions

/// Creates a search
///
/// \param searchRoot
/// \param pattern
/// \param flags
/// \param hSrch
/// \return 
/// \retval

public recls_rc_t Search_Create(in char[] searchRoot, in char[] pattern, in int flags, out hrecls_t hSrch)
{
    // Always ensure that the semantics of null reflect those of the core API
    char    *root_ptr;
    char    *pattern_ptr;

    if(null !is searchRoot)
    {
        root_ptr = toStringz(searchRoot);
    }
    else
    {
        root_ptr = null;
    }

    if(null !is pattern)
    {
        pattern_ptr = toStringz(pattern);
    }
    else
    {
        pattern_ptr = null;
    }

    return Recls_Search(root_ptr, pattern_ptr, flags, &hSrch);
}

public recls_rc_t Search_CreateFtp(in char[] host, in char[] username, in char[] password, in char[] searchRoot, in char[] pattern, in int flags, out hrecls_t hSrch)
{
    // Always ensure that the semantics of null reflect those of the core API
    char    *host_ptr       =   (null is host)         ? null : toStringz(host);
    char    *username_ptr   =   (null is username)     ? null : toStringz(username);
    char    *password_ptr   =   (null is password)     ? null : toStringz(password);
    char    *root_ptr       =   (null is searchRoot)   ? null : toStringz(searchRoot);
    char    *pattern_ptr    =   (null is pattern)      ? null : toStringz(pattern);

    return Recls_SearchFtp(host_ptr, username_ptr, password_ptr, root_ptr, pattern_ptr, flags, &hSrch);
}

/+
private extern(Windows) int process_fn(recls_info_t entry, recls_process_fn_param_t p)
{
    return dg(Entry._make_Entry(entry), p);
}

public recls_rc_t Search_Process(   in char[]                                                       searchRoot
                                ,   in char[]                                                       pattern
                                ,   in int                                                          flags
                                ,   int delegate(in Entry entry, recls_process_fn_param_t param)    dg
                                ,   recls_process_fn_param_t                                        param)
{
/*     extern(Windows) int process_fn(recls_info_t entry, recls_process_fn_param_t p)
    {
        return dg(Entry._make_Entry(entry), p);
    }
 */
    return Recls_SearchProcess(searchRoot, pattern, flags, process_fn, param);
}
+/

/// Advances the given search to the next position
///
/// \param hSrch handle identifying the search
/// \return return code indicating status of the operation
/// \return RECLS_
public recls_rc_t Search_GetNext(in hrecls_t hSrch)
{
    return Recls_GetNext(hSrch);
}

/// Closes the given search
///
/// \param hSrch handle identifying the search
public void Search_Close(inout hrecls_t hSrch)
{
    Recls_SearchClose(hSrch);

    hSrch = null;
}

public recls_rc_t Search_GetEntry(in hrecls_t hSrch, out recls_info_t entry)
{
    return Recls_GetDetails(hSrch, entry);
}

public recls_rc_t Search_GetNextEntry(in hrecls_t hSrch, out recls_info_t entry)
{
    return Recls_GetNextDetails(hSrch, entry);
}

public void Search_CloseEntry(inout recls_info_t entry)
{
    Recls_CloseDetails(entry);

    entry = null;
}

public recls_info_t Search_CopyEntry(in recls_info_t entry)
{
    recls_info_t    copy;

    if(RECLS_FAILED(Recls_CopyDetails(entry, &copy)))
    {
        copy = null;
    }

    return copy;
}

public recls_rc_t Search_OutstandingDetails(in hrecls_t hSrch, out recls_uint32_t count)
{
    return Recls_OutstandingDetails(hSrch, count);
}

public recls_rc_t Search_GetLastError(in hrecls_t hSrch)
{
    return Recls_GetLastError(hSrch);
}

public char[] Search_GetErrorString(in recls_rc_t rc)
{
    uint    cch =   Recls_GetErrorString(rc, null, 0);
    char[]  err =   new char[cch];

    cch = Recls_GetErrorString(rc, err, err.length);

    assert(cch <= err.length);

    return MBStoUTF8(err);
}

public char[] Search_GetEntryPath(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint    cch     =   Recls_GetPathProperty(entry, null, 0);
    char[]  path    =   new char[cch];

    cch = Recls_GetPathProperty(entry, path, path.length);

    assert(cch <= path.length);

    path = MBStoUTF8(path);

//  printf("[%.*s\n", path);
//  std.stdio.writef(path);

    return path;
}

public char[] Search_GetEntrySearchRelativePath(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint    cch     =   Recls_GetSearchRelativePathProperty(entry, null, 0);
    char[]  path    =   new char[cch];

    cch = Recls_GetSearchRelativePathProperty(entry, path, path.length);

    assert(cch <= path.length);

    path = MBStoUTF8(path);

//  printf("[%.*s\n", path);
//  std.stdio.writef(path);

    return path;
}

version(Windows)
{
public char Search_GetEntryDrive(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    char    chDrive;

    return (Recls_GetDriveProperty(entry, chDrive), chDrive);
}
} // version Windows

public char[] Search_GetEntryDirectory(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint    cch =   Recls_GetDirectoryProperty(entry, null, 0);
    char[]  str =   new char[cch];

    cch = Recls_GetDirectoryProperty(entry, str, str.length);

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

public char[] Search_GetEntryDirectoryPath(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint    cch =   Recls_GetDirectoryPathProperty(entry, null, 0);
    char[]  str =   new char[cch];

    cch = Recls_GetDirectoryPathProperty(entry, str, str.length);

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

public char[] Search_GetEntrySearchDirectory(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint    cch =   Recls_GetSearchDirectoryProperty(entry, null, 0);
    char[]  str =   new char[cch];

    cch = Recls_GetSearchDirectoryProperty(entry, str, str.length);

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

public char[] Search_GetEntryUNCDrive(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint    cch =   Recls_GetUNCDriveProperty(entry, null, 0);
    char[]  str =   new char[cch];

    cch = Recls_GetUNCDriveProperty(entry, str, str.length);

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

public char[] Search_GetEntryFile(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint    cch =   Recls_GetFileProperty(entry, null, 0);
    char[]  str =   new char[cch];

    cch = Recls_GetFileProperty(entry, str, str.length);

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

public char[] Search_GetEntryShortFile(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint    cch =   Recls_GetShortFileProperty(entry, null, 0);
    char[]  str =   new char[cch];

    cch = Recls_GetShortFileProperty(entry, str, str.length);

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

public char[] Search_GetEntryFileName(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint    cch =   Recls_GetFileNameProperty(entry, null, 0);
    char[]  str =   new char[cch];

    cch = Recls_GetFileNameProperty(entry, str, str.length);

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

public char[] Search_GetEntryFileExt(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint    cch =   Recls_GetFileExtProperty(entry, null, 0);
    char[]  str =   new char[cch];

    cch = Recls_GetFileExtProperty(entry, str, str.length);

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

public char[][] Search_GetEntryDirectoryParts(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    uint        cParts  =   Recls_GetDirectoryPartProperty(entry, -1, null, 0);
    char[][]    parts   =   new char[][cParts];

    for(int i = 0; i < cParts; ++i)
    {
        uint    cch =   Recls_GetDirectoryPartProperty(entry, i, null, 0);
        char[]  str =   new char[cch];

        cch = Recls_GetDirectoryPartProperty(entry, i, str, str.length);

        assert(cch <= str.length);

        parts[i] = MBStoUTF8(str);
    }

    return parts;
}

public boolean Search_IsEntryReadOnly(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    return Recls_IsFileReadOnly(entry);
}

public boolean Search_IsEntryDirectory(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    return Recls_IsFileDirectory(entry);
}

public boolean Search_IsEntryLink(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    return Recls_IsFileLink(entry);
}

public boolean Search_IsEntryUNC(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    return Recls_IsFileUNC(entry);
}

public recls_filesize_t Search_GetEntrySize(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    recls_filesize_t    size;

    return (Recls_GetSizeProperty(entry, &size), size);
}

public d_time Search_GetEntryCreationTime(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    return recls_time2d_time(Recls_GetCreationTime(entry));
}

public d_time Search_GetEntryModificationTime(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    return recls_time2d_time(Recls_GetModificationTime(entry));
}

public d_time Search_GetEntryLastAccessTime(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    return recls_time2d_time(Recls_GetLastAccessTime(entry));
}

public d_time Search_GetEntryLastStatusChangeTime(in recls_info_t entry)
in
{
    assert(null !is entry);
}
body
{
    return recls_time2d_time(Recls_GetLastStatusChangeTime(entry));
}

public char[][] getRoots()
{
    recls_root_t    root_structs[26];
    uint            cRoots  =   Recls_GetRoots(&root_structs[0], root_structs.length);
    char[][]        roots   =   new char[][cRoots];

    for(uint i = 0; i < cRoots; ++i)
    {
        roots[i] = MBStoUTF8(root_structs[i].name);
    }

    return roots;
}

public char[][] getSelectedRoots(uint flags)
{
    recls_root_t    root_structs[26];
    uint            cRoots  =   Recls_GetSelectedRoots(&root_structs[0], root_structs.length, flags);
    char[][]        roots   =   new char[][cRoots];

    for(uint i = 0; i < cRoots; ++i)
    {
        roots[i] = MBStoUTF8(root_structs[i].name);
    }

    return roots;
}

public char[] getPathNameSeparator()
{
    char    *s  =   Recls_GetPathNameSeparator();

    return MBSztoUTF8(s);
}

public char[] getPathSeparator()
{
    char    *s  =   Recls_GetPathSeparator();

    return MBSztoUTF8(s);
}

public char[] wildcardsAll()
{
    char    *s  =   Recls_GetWildcardsAll();

    return MBSztoUTF8(s);
}

private boolean isDirectoryEmpty_(char[] dir)
{
    FileSearch  search  =       new FileSearch(dir, wildcardsAll(), RECLS_FLAG.RECLS_F_FILES | RECLS_FLAG.RECLS_F_DIRECTORIES | RECLS_FLAG.RECLS_F_DETAILS_LATER);

    foreach(Entry entry; search)
    {
        return false;
    }

    return true;
}

/// \brief This function returns indicates whether the 
/// \note It throws a ReclsException if the given path is not actually a directory
public boolean isDirectoryEmpty(char[] dir)
in
{
    assert(null !is dir);
}
body
{
    if(!isRoot_(dir))
    {
        if(!stat(dir, 0).isDirectory())
        {
            throw new ReclsException("\"" ~ dir ~ "\" is not a directory");
        }
    }

    return isDirectoryEmpty_(dir);
}

/// \brief This function returns indicates whether the 
/// \note It throws a ReclsException if the given entry is not actually a directory
public boolean isDirectoryEmpty(Entry dir)
in
{
    assert(null !is dir);
}
body
{
    if(!dir.isDirectory())
    {
        throw new ReclsException("\"" ~ dir.path ~ "\" is not a directory");
    }
    else
    {
        return isDirectoryEmpty_(dir.path);
    }
}

private recls_filesize_t calcDirectorySize_(char[] dir, boolean bRecurse)
{
    recls_filesize_t    total   =   0;
    uint                flags   =   RECLS_FLAG.RECLS_F_FILES | RECLS_FLAG.RECLS_F_DIRECTORIES;
    
    if(bRecurse)
    {
        flags |=  RECLS_FLAG.RECLS_F_RECURSIVE;
    }

    foreach(Entry entry; new FileSearch(dir, "*.*", flags))
    {
        total += entry.fileSize;
    }

    return total;
}


public recls_filesize_t calcDirectorySize(char[] dir, boolean bRecurse)
in
{
    assert(null !is dir);
}
body
{
    if(!isRoot_(dir))
    {
        if(!stat(dir, 0).isDirectory())
        {
            throw new ReclsException("\"" ~ dir ~ "\" is not a directory");
        }
    }

    return calcDirectorySize_(dir, bRecurse);
}

public recls_filesize_t calcDirectorySize(Entry dir, boolean bRecurse)
in
{
    assert(null !is dir);
}
body
{
    if(!dir.isDirectory())
    {
        throw new ReclsException("\"" ~ dir.path ~ "\" is not a directory");
    }
    else
    {
        return calcDirectorySize_(dir.path, bRecurse);
    }
}

public Entry stat(char[] path, in uint flags)
in
{
    assert(null !is path);
}
body
{
    recls_info_t    info;
    recls_rc_t      rc  =   Recls_Stat(toStringz(path), flags, &info);

    if(RECLS_FAILED(rc))
    {
        if(RECLS_RC_NO_MORE_DATA == rc)
        {
            return null;
        }
        else
        {
            throw new ReclsException(rc);
        }
    }
    else
    {
        return new Entry(info);
    }
}

public Entry stat(char[] path)
{
	return stat(path, 0);
}

public char[] combinePaths(char[] path1, char[] path2)
in
{
	assert(null !is path1 || null !is path2);
}
body
{
    uint    cch =   Recls_CombinePaths(path1, path2, null, 0);
    char[]  str =   new char[cch];

    cch = Recls_CombinePaths(path1, path2, str, str.length);

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

public char[] deriveRelativePath(char[] origin, char[] target)
in
{
	assert(null !is origin);
	assert(null !is target);
}
body
{
    uint    cch =   Recls_DeriveRelativePath(origin, target, null, 0);
    char[]  str =   new char[cch];

    cch = Recls_DeriveRelativePath(origin, target, str, str.length);

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

public char[] squeezePath(char[] path, uint width)
in
{
	assert(null !is path);
}
body
{
    uint    cch =   Recls_SqueezePath(path, null, width);
    char[]  str =   new char[cch];

    cch = Recls_SqueezePath(path, str, str.length);

	str = str[0 .. cch];

    assert(cch <= str.length);

    return MBStoUTF8(str);
}

////////////////////////////////////////////////////////////////////////////////
// Classes

/// Exception class used by the recls module
/// \ingroup group_recls_d
public class ReclsException
    : Exception
{
public:
    this(recls_rc_t rc)
    {
        super(rc_to_string(rc));
    }
    this(char[] message, recls_rc_t rc)
    {
        super(message ~ rc_to_string(rc));
    }
    this(char[] message)
    {
        super(message);
    }

private:
    static char[] rc_to_string(recls_rc_t rc)
    {
        int     cch =   Recls_GetErrorString(rc, null, 0);
        char[]  s   =   new char[cch];

        Recls_GetErrorString(rc, &s[0], s.length);

        return s;
    }
}

/// Base class for FileSearch and FtpSearch
/// \ingroup group_recls_d
public /* abstract */ class Search
{
public:
    int opApply(int delegate(inout Entry entry) dg)
    {
        int             result  =   0;
        hrecls_t        hSrch;

        recls_rc_t      rc      =   createSearch(hSrch);

        do
        {
            recls_info_t    entry;

            if(RECLS_FAILED(rc))
            {
                if(RECLS_RC_NO_MORE_DATA != rc)
                {
                    // throw new ReclsException("Search continuation failed", rc);
                }

                result = 1;
            }
            else
            {
                rc = Search_GetEntry(hSrch, entry);

                if(RECLS_FAILED(rc))
                {
                    if(RECLS_RC_NO_MORE_DATA != rc)
                    {
                        // throw new ReclsException("Search continuation failed", rc);
                    }

                    result = 1;
                }
                else
                {
                    try
                    {
                        Entry   e = Entry._make_Entry(entry);

                        result = dg(e);
                    }
                    finally
                    {
                        Search_CloseEntry(entry);
                    }
                }

                rc = Search_GetNextEntry(hSrch, entry);
            }

        } while(0 == result);

        return result;
    }

protected:
    abstract recls_rc_t createSearch(out hrecls_t hSrch)
    {
        assert(0);

        return cast(recls_rc_t)(-1);
    }
}

/// Conducts a file-system search
/// \ingroup group_recls_d
public class FileSearch
    : public Search
{
/// \name Construction
public:
    /// Create a search object with the given searchRoot, pattern and flags
    ///
    /// \param searchRoot The root directory of the search. If null, or the empty string, the current directory is assumed
    /// \param pattern The search pattern. If null, or the empty string, all entries are returned
    /// \param flags A combination of RECLS_FLAG flags which moderate the search
    this(in char[] searchRoot, in char[] pattern, in uint flags)
    {
        m_searchRoot    =   searchRoot;
        m_pattern       =   pattern;
        m_flags         =   flags;
    }

version(ReclsHasEnumerator)
{
/// \name Types
public:
    class Enumerator
    {
    private:
        this(hrecls_t hSrch, recls_rc_t lastError)
        {
            m_hSrch     =   hSrch;
            m_lastError =   lastError;
        }
    public:
        ~this()
        {
            if(null != m_hSrch)
            {
                Search_Close(m_hSrch);
            }
        }

    public:
        boolean hasEntry()
        {
            return RECLS_SUCCEEDED(m_lastError);
        }

        Entry   currentEntry()
        in
        {
            assert(null !is m_hSrch);
        }
        body
        {
            recls_info_t    entry;
            recls_rc_t      rc  =   Search_GetEntry(m_hSrch, entry);

            m_lastError = rc;

            try
            {
                return Entry._make_Entry(entry);
            }
            finally
            {
                Search_CloseEntry(entry);
            }
        }

        recls_rc_t lastError()
        {
            return m_lastError;
        }

    public:
        boolean getNextEntry()
        in
        {
            assert(null != m_hSrch);
        }
        body
        {
            recls_rc_t  rc  =   Search_GetNext(m_hSrch);

            m_lastError = rc;

            if(RECLS_FAILED(rc))
            {
                Search_Close(m_hSrch);

                if(RECLS_RC_NO_MORE_DATA != rc)
                {
                    // throw new ReclsException("Search continuation failed", rc);
                }
            }

            return RECLS_SUCCEEDED(rc);
        }

    /// Members
    private:
        hrecls_t    m_hSrch;    // NOTE THAT D DOES STRONG TYPEDEFS (see true-typedefs)
        recls_rc_t  m_lastError;
    }

/// Operations
public:
    Enumerator enumerate()
    {
        hrecls_t    hSrch;
        recls_rc_t  rc  =   Search_Create(m_searchRoot, m_pattern, m_flags, hSrch);

        try
        {
            return new Enumerator(hSrch, rc);
        }
        catch(Exception x)
        {
            Search_Close(hSrch);

            throw x;
        }
    }
} // version(ReclsHasEnumerator)

protected:
    override recls_rc_t createSearch(out hrecls_t hSrch)
    {
        return Search_Create(m_searchRoot, m_pattern, m_flags, hSrch);
    }

/// Members
private:
    char[]  m_searchRoot;
    char[]  m_pattern;
    uint    m_flags;
}

version(Windows)
{

/// Conducts a search of an FTP host
/// \ingroup group_recls_d
public class FtpSearch
    : public Search
{
/// \name Construction
public:
    /// Create a search object with the given searchRoot, pattern and flags
    ///
    /// \param searchRoot The root directory of the search. If null, or the empty string, the current directory is assumed
    /// \param pattern The search pattern. If null, or the empty string, all entries are returned
    /// \param flags A combination of RECLS_FLAG flags which moderate the search
    this(in char[] host, in char[] username, in char[] password, in char[] searchRoot, in char[] pattern, in uint flags)
    {
        m_host          =   host;
        m_username      =   username;
        m_password      =   password;
        m_searchRoot    =   searchRoot;
        m_pattern       =   pattern;
        m_flags         =   flags;
    }

protected:
    override recls_rc_t createSearch(out hrecls_t hSrch)
    {
        return Search_CreateFtp(m_host, m_username, m_password, m_searchRoot, m_pattern, m_flags, hSrch);
    }

/// Members
private:
    char[]  m_host;
    char[]  m_username;
    char[]  m_password;
    char[]  m_searchRoot;
    char[]  m_pattern;
    uint    m_flags;
}

} // version(Windows)

/* public class Boolean
{
public:
    this(boolean value)
    {
        m_value = value;
    }

    op()
    {
        return m_value != 0;
    }

private:
    boolean m_value;
}
 */

/// Represents a search entry
/// \ingroup group_recls_d
public class Entry
{
    invariant
    {
        if(null != m_entry)
        {
            // Now do all the checks to verify that the various components of the path are valid

            // Since we cannot call member functions (as that would end up in recursion)
            // the only thing we can do is to "test" the validity of the entry handle, so
            // we just add a reference, and then release it
            recls_info_t    entry   =   Search_CopyEntry(m_entry);

            assert(null !is entry);

            Recls_CloseDetails(entry);
        }
    }

private:
    /// This is necessary, because DMD 0.73 generates code
    /// that goes into an infinite loop when creating an
    /// Entry instance with a non-null entry
    static Entry _make_Entry(recls_info_t entry)
    {
        recls_info_t    copy    =   Search_CopyEntry(entry);
        Entry           e       =   null;

        try
        {
            e = new Entry(null);

            e.m_entry = entry;
        }
        catch(Exception x)
        {
            Search_CloseEntry(copy);

            throw x;
        }

        return e;
    }

    this(recls_info_t entry)
    {
        m_entry = entry;
    }
    ~this()
    {
        if(null !is m_entry)
        {
            Search_CloseEntry(m_entry);
        }
    }

public:
    /// The full path of the entry
    ///
    /// \note For "/usr/include/recls/recls_assert.h" this would yield "/usr/include/recls/recls_assert.h"
    char[]              getPath()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryPath(m_entry);
    }

    /// The root directory of the search
    char[]              getSearchDirectory()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntrySearchDirectory(m_entry);
    }

    /// The path of the entry, relative to the search directory
    char[]              getSearchRelativePath()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntrySearchRelativePath(m_entry);
    }

    /// The full path of the entry
    ///
    /// \note For "/usr/include/recls/recls_assert.h" this would yield "/usr/include/recls/recls_assert.h"
    char[]              path()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryPath(m_entry);
    }
version(Windows)
{
    /// The drive component of the entry's path
    ///
    /// \note For "H:\Dev\include\recls\recls_assert.h" this would yield 'H'
    char                drive()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryDrive(m_entry);
    }
} // version(Windows)
    /// The directory component of the entry's path
    ///
    /// \note For "/usr/include/recls/recls_assert.h" this would yield "/usr/include/recls/"
    char[]              directory()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryDirectory(m_entry);
    }
    /// The full location component of the entry's path.
    ///
    /// \note This is everything before the filename+fileext. On Win32 systems for "H:\Dev\include\recls\recls_assert.h" this would yield "H:\Dev\include\recls\"
    char[]              directoryPath()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryDirectoryPath(m_entry);
    }
    /// The part of the path, if any, that is UNC
    ///
    /// \note This is only supported on Win32 systems. Other systems will always yield the empty string
    char[]              UNCDrive()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryUNCDrive(m_entry);
    }
    /// An array of strings representing the parts of the Directory property
    ///
    /// \note For "/usr/include/recls/recls_assert.h" this would yield [ "/", "usr/", "include/", "recls/"]
    char[][]            directoryParts()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryDirectoryParts(m_entry);
    }
    /// The file component of the entry's path
    ///
    /// \note For "/usr/include/recls/recls_assert.h" this would yield "recls_assert.h"
    char[]              file()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryFile(m_entry);
    }
    /// The short equivalent of the entry's File property
    ///
    /// \note On Win32 systems, this is the 8.3 form, e.g. "recls_~1.h". On other systems this is identical to the File property
    char[]              shortFile()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryShortFile(m_entry);
    }
    /// The file name component of the entry's path
    ///
    /// \note For "/usr/include/recls/recls_assert.h" this would yield "recls_assert"
    char[]              fileName()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryFileName(m_entry);
    }
    /// The file extension component of the entry's path
    ///
    /// \note For "/usr/include/recls/recls_assert.h" this would yield "h"
    char[]              fileExt()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryFileExt(m_entry);
    }

    /// The time the entry was created
    d_time              creationTime()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryCreationTime(m_entry);
    }
    /// The time the entry was last modified
    d_time              modificationTime()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryModificationTime(m_entry);
    }
    /// The time the entry was last accessed
    d_time              lastAccessTime()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryLastAccessTime(m_entry);
    }
    /// The time the entry's last status changed
    d_time              lastStatusChangeTime()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntryLastStatusChangeTime(m_entry);
    }

    /// The size of the entry
    recls_filesize_t    fileSize()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_GetEntrySize(m_entry);
    }

    /// Indicates whether the entry is read-only
    boolean             isReadOnly()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_IsEntryReadOnly(m_entry);
    }
    /// Indicates whether the entry is a directory
    boolean             isDirectory()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_IsEntryDirectory(m_entry);
    }
    /// Indicates whether the entry is a link
    boolean             isLink()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_IsEntryLink(m_entry);
    }
    /// Indicates whether the entry's path is UNC
    boolean             isUNC()
    in
    {
        assert(null !is m_entry);
    }
    body
    {
        return Search_IsEntryUNC(m_entry);
    }

/// Members
private:
    recls_info_t    m_entry;
}

////////////////////////////////////////////////////////////////////////////////

unittest
{
//  printf("std/recls.d unittest\n");

    FileSearch  search  =   new FileSearch(".", wildcardsAll(), RECLS_FLAG.RECLS_F_RECURSIVE);

    foreach(Entry entry; search)
    {
/+
        printf("std/recls.d: entry.path: %.*s\n", entry.path());
        printf("std/recls.d: entry.directory: %.*s\n", entry.directory());
        printf("std/recls.d: entry.directoryPath: %.*s\n", entry.directoryPath());
        printf("std/recls.d: entry.UNCDrive: %.*s\n", entry.UNCDrive());
        printf("std/recls.d: entry.file: %.*s\n", entry.file());
        printf("std/recls.d: entry.shortFile: %.*s\n", entry.shortFile());
        printf("std/recls.d: entry.fileExt: %.*s\n", entry.fileExt());
        printf("std/recls.d: entry.creationTime: %.*s\n", std.date.toString(entry.creationTime()));
        printf("std/recls.d: entry.modificationTime: %.*s\n", std.date.toString(entry.modificationTime()));
        printf("std/recls.d: entry.lastAccessTime: %.*s\n", std.date.toString(entry.lastAccessTime()));
        printf("std/recls.d: entry.lastStatusChangeTime: %.*s\n", std.date.toString(entry.lastStatusChangeTime()));
        printf("std/recls.d: entry.fileSize: %lu\n", entry.fileSize());
/+
        printf("std/recls.d: entry.path: %.*s\n", entry.path());
        printf("std/recls.d: entry.path: %.*s\n", entry.path());
        printf("std/recls.d: entry.path: %.*s\n", entry.path());
+/
+/

        entry.path();
version(Windows)
{
        entry.drive();
} // version(Windows)
        entry.directory();
        entry.directoryPath();
        entry.UNCDrive();
        entry.directoryParts();
        entry.file();
        entry.shortFile();
        entry.fileName();
        entry.fileExt();
        entry.creationTime();
        entry.modificationTime();
        entry.lastAccessTime();
        entry.lastStatusChangeTime();
        entry.fileSize();
        entry.isReadOnly();
        entry.isDirectory();
        entry.isLink();
        entry.isUNC();
    }

//  printf("std/recls.d unittest = complete\n");
}

version(MainTest)
{
    private char[] toString(boolean b)
    {
        return b ? "true" : "false";
    }

    private char[] stuff_(int off)
    {
        char[]  s   =   new char[off];

        if(0 != off)
        {
            std.c.string.memset(&s[0], ' ', off);
        }

        return s;
    }

    private void dumpEntry(Entry entry)
    {
        printf("  %.*s\n", entry.path);
        printf("  %.*s\n", entry.directoryPath);
        char    drive       =   '\0';
version(Windows)
{
        drive = entry.drive;
} // version(Windows)
        char[]  uncDrive    =   entry.UNCDrive;
        int     off         =   0;
        if(0 != uncDrive.length)
        {
            printf("  %.*s\n", uncDrive);
            off = uncDrive.length;
        }
        else if('\0' != drive)
        {
            printf("  %c\n", drive);
            off = 2;
        }
        else
        {
        }

        printf("  %.*s%.*s\n", stuff_(off), entry.directory);
        if(0 == entry.directoryParts.length)
        {
            off += entry.directory.length;
        }
        else
        {
            foreach(char[] part; entry.directoryParts)
            {
                printf("  %.*s%.*s\n", stuff_(off), part);

                off += part.length;
            }
        }
        printf("  %.*s%.*s\n", stuff_(off), entry.file);
        printf("  %.*s%.*s\n", stuff_(off), entry.fileName);
        off += entry.fileName.length + 1;
        printf("  %.*s%.*s\n", stuff_(off), entry.fileExt);
        printf("  %llu bytes\n", entry.fileSize);
        if(entry.isReadOnly)
        {
            printf("  - read-only\n");
        }
        else
        {
            printf("  - writeable\n");
        }
        if(entry.isDirectory)
        {
            printf("  - directory\n");
        }
        else
        {
            printf("  - file\n");
        }
        if(entry.isLink)
        {
            printf("  - link\n");
        }
        if(entry.isUNC)
        {
            printf("  - unc\n");
        }
    }

    int main()
    {
        printf("recls test\n\n");

        printf("1.a Enumerating roots, calculating size and determining emptiness:\n\n");
        try
        {
            foreach(char[] root; getRoots())
            {
                printf("  %.*s\n", root);
                printf("    empty: %.*s\n", toString(isDirectoryEmpty(root)));
                printf("    size:  %llu\n", calcDirectorySize(root, false));
            }
        }
        catch(Exception x)
        {
            printf("  Error: %.*s\n", x.toString());
        }
        printf("\n");

        printf("1.b Enumerating selected roots, calculating size and determining emptiness:\n\n");
        try
        {
            foreach(char[] root; getSelectedRoots(RECLS_ROOTS_FLAG.NETWORK_DRIVES))
            {
                printf("  %.*s\n", root);
                printf("    empty: %.*s\n", toString(isDirectoryEmpty(root)));
                printf("    size:  %llu\n", calcDirectorySize(root, false));
            }
        }
        catch(Exception x)
        {
            printf("  Error: %.*s\n", x.toString());
        }
        printf("\n");

        printf("2. Enumerating contents of current directory (non-recursively, with directory parts):\n\n");
        try
        {
            foreach(Entry entry; new FileSearch(".", wildcardsAll(), RECLS_FLAG.FILES | RECLS_FLAG.DIRECTORIES | RECLS_FLAG.DIRECTORY_PARTS))
//          foreach(Entry entry; new FileSearch("\\\\jau-xp\\H$\\dev\\bin", wildcardsAll(), RECLS_FLAG.FILES | RECLS_FLAG.DIRECTORIES | RECLS_FLAG.DIRECTORY_PARTS))
            {
                dumpEntry(entry);
                printf("\n");
            }
        }
        catch(Exception x)
        {
            printf("  Error: %.*s\n", x.toString());
        }
        printf("\n");

        printf("3. Using stat() on files in current directory (without directory parts):\n\n");
        try
        {
            foreach(Entry entry; new FileSearch(".", wildcardsAll(), RECLS_FLAG.FILES))
            {
                dumpEntry(stat(entry.path, 0));
            }
        }
        catch(Exception x)
        {
            printf("  Error: %.*s\n", x.toString());
        }
        printf("\n");

version(Windows)
{
        printf("4. Enumerating contents of ftp://ftp.digitalmars.com/ (recursively, with directory parts):\n\n");
        try
        {
            foreach(Entry entry; new FtpSearch("ftp.digitalmars.com", "anonymous", "", "/", wildcardsAll(), RECLS_FLAG.FILES | RECLS_FLAG.DIRECTORIES | RECLS_FLAG.RECURSIVE | RECLS_FLAG.DIRECTORY_PARTS))
            {
                dumpEntry(entry);
            }
        }
        catch(Exception x)
        {
            printf("  Error: %.*s\n", x.toString());
        }
        printf("\n");
} // version(Windows)

        printf("5. Using combinePaths(), deriveRelativePath() and squeezePath():\n\n");
        try
        {
			char[]	res1	=	combinePaths(stat(".").path, "sub1");

			printf("combined path: %.*s\n", res1);

			char[]	res2	=	deriveRelativePath("/usr/include/recls/", "/usr/lib/recls");

			printf("relative path: %.*s\n", res2);

			for(int i = 0; i < res1.length + 10; ++i)
			{
				char[]	res3	=	squeezePath(res1, i);

				printf("  squeezing into %2u characters: %.*s\n", i, res3);
			}
        }
        catch(Exception x)
        {
            printf("  Error: %.*s\n", x.toString());
        }
        printf("\n");

        return 0;
    }
}

////////////////////////////////////////////////////////////////////////////////
