/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls.ruby.c
 *
 * Purpose:     Implementation of the recls-Ruby mapping
 *
 * Created      14th March 2004
 * Updated:     16th April 2007
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2004-2007, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* \file recls.ruby.c
 *
 * \brief Implementation of the recls-Ruby mapping
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Version information
 */

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_C_RECLS_RUBY_MAJOR     1
# define RECLS_VER_C_RECLS_RUBY_MINOR     9
# define RECLS_VER_C_RECLS_RUBY_REVISION  2
# define RECLS_VER_C_RECLS_RUBY_EDIT      38
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <ruby.h>
#include <recls/recls.h>
#include <recls/assert.h>
#include <recls/minmax.h>

#define STLSOFT_FORCE_C_COMPILATION_IN_BETA

#include <stlsoft/stlsoft.h>    /* If the compiler cannot find this, you are not using STLSoft 1.9.1 or later, as required. */
#if !defined(_STLSOFT_VER_1_9_1)
# error STLSoft version 1.9.1 or greater required. (www.stlsoft.org/downloads.html)
#endif /* STLSoft version */

#if defined(RECLS_PLATFORM_IS_WIN32)
# include <winstl/time/conversion_functions.h>
#endif /* RECLS_PLATFORM_IS_WIN32 */

#include <stdarg.h>

/* /////////////////////////////////////////////////////////////////////////////
 * Ruby version compatibility
 */

#ifndef StringValuePtr
# define StringValuePtr(x)  STR2CSTR(x)
#endif /* !StringValuePtr */

/* /////////////////////////////////////////////////////////////////////////////
 * Compiler warnings
 */

#ifdef _MSC_VER
# pragma warning(1 : 4100)
# pragma warning(1 : 4101)
# pragma warning(1 : 4130)
# pragma warning(1 : 4131)
# pragma warning(1 : 4132)
# pragma warning(1 : 4189)
# pragma warning(1 : 4209)
# pragma warning(1 : 4213)
# pragma warning(1 : 4214)
# pragma warning(1 : 4220)
# pragma warning(disable : 4127)
# pragma warning(disable : 4505)
#endif /* _MSC_VER */

/* /////////////////////////////////////////////////////////////////////////////
 * Helper functions & macros
 */

/* Extract from STLSoft's stlsoft.h */

#if !defined(STLSOFT_SUPPRESS_UNUSED)
# define STLSOFT_SUPPRESS_UNUSED(x)     ((void)x)
#endif /* !STLSOFT_SUPPRESS_UNUSED */

/* Extract from Synesis Software's SLBase.h */
#ifndef RECLS_NUM_ELEMENTS
# define RECLS_NUM_ELEMENTS(x)          (sizeof(x) / sizeof(0[(x)]))
#endif /* RECLS_NUM_ELEMENTS */

#define __synsoft_stringize(x)          #x
#define __synsoft_stringize_symbol(x)   __synsoft_stringize(x) /* x is translated prior to stringizing */

/* Extract from Synesis Software's MLLimits.h */
#if (   defined(__DMC__) || \
        defined(__COMO__) || \
        defined(__GNUC__) || \
        defined(__MWERKS__))
# define __SYNSOFT_GEN_S64BIT_SUFFIX(i)         (i ## LL)
# define __SYNSOFT_GEN_U64BIT_SUFFIX(i)         (i ## ULL)
#elif ( defined(__BORLANDC__) || \
        defined(__INTEL_COMPILER) || \
        defined(_MSC_VER) || \
        defined(__VECTORC) || \
        defined(__WATCOMC__))
# define __SYNSOFT_GEN_S64BIT_SUFFIX(i)         (i ## L)
# define __SYNSOFT_GEN_U64BIT_SUFFIX(i)         (i ## UL)
#else
# error Compiler not discriminated
#endif /* __SYNSOFT_DVS_COMPILER_GNUC || __SYNSOFT_DVS_COMPILER_DECCXX */

/* /////////////////////////////////////////////////////////////////////////////
 * Simple debugging
 */

/* #undef _MSC_VER */
/* #undef WIN32 */
/* #define DEBUG */

static void TraceToDebugger(char const *fmt, ...)
{
    va_list args;
#ifdef WIN32
    char    sz[2048];
#endif /* ? WIN32 */

    va_start(args, fmt);

#ifdef WIN32
# ifdef _MSC_VER
    _vsnprintf(sz, RECLS_NUM_ELEMENTS(sz) - 1, fmt, args);
# else /* ? compiler */
    vsprintf(sz, fmt, args);
# endif /* compiler */
    OutputDebugStringA(sz);
#else /* ? OS */
    vfprintf(stdout, fmt, args);
#endif /* OS */

    va_end(args);
}

#if !defined(NDEBUG) && \
    (   defined(_DEBUG) || \
        defined(DEBUG))
# define TRACETODEBUGGER    TraceToDebugger
#else
static void Trace_Stub(char const *fmt, ...)
{
    STLSOFT_SUPPRESS_UNUSED(fmt);
}
# define TRACETODEBUGGER    (1) ? ((void)0) : Trace_Stub
#endif /* debug */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes & Modules
 */

static VALUE    mRecls;             /* The Recls module:            ::Recls                 */
static VALUE    cFileSearch;        /* The FileSearch class:        ::Recls::FileSearch     */
static VALUE    cFtpSearch;         /* The FtpSearch class:         ::Recls::FtpSearch      */
static VALUE    cEntry;             /* The Entry class:             ::Recls::Entry          */
static VALUE    eReclsException;    /* The recls exception class:   ::Recls::ReclsException */

static VALUE    s_currentDirectory; /* The current directory                                */
static VALUE    s_wildcardsAll;     /* The 'search everything' pattern                      */

/* /////////////////////////////////////////////////////////////////////////////
 * Helper functions & macros
 */

/** Constructs a Ruby string from a string range. */
static VALUE rb_str_from_ptrs(char const *begin, char const *end)
{
    RECLS_ASSERT(begin <= end);

    return rb_str_new(begin, end - begin);
}

/** Constructs a Ruby string from a recls_strptrs_t structure. */
static VALUE rb_str_from_strptrs(const struct recls_strptrs_t *ptrs)
{
    return rb_str_from_ptrs(ptrs->begin, ptrs->end);
}

/* /////////////////////////////////////////////////////////////////////////////
 * Entry
 */

/** Free the resources associated with the entry */
static void Entry_free(void *p)
{
    recls_info_t    entry = (recls_info_t)(p);

    TRACETODEBUGGER("Closing entry (0x%p, \"%s\")\n", entry, entry->path);

    Recls_CloseDetails(entry);
}

/** Translate the self value into a Entry structure. */
static recls_info_t Entry__entry_handle_get_(VALUE self)
{
    recls_info_t    entry;

    Data_Get_Struct(self, struct recls_fileinfo_t const, entry);

    return entry;
}

/** Retrieve the string value of one of the strptrs, via its offset within the entry structure. */
static VALUE Entry_strptr_get_(VALUE self, size_t offset)
{
    void                    *pvEntry;
    struct recls_strptrs_t  *ptrs;

    Data_Get_Struct(self, void*, pvEntry);
    ptrs = (struct recls_strptrs_t*)((char const*)(pvEntry) + offset);

    return rb_str_from_strptrs(ptrs);
}

/** Retrieves the time value of one of the attributes, via the requisite API functions. */
static VALUE Entry_time_get_(VALUE self, recls_time_t (RECLS_CALLCONV_DEFAULT *pfn)(recls_info_t))
{
    recls_info_t    entry   =   Entry__entry_handle_get_(self);
    recls_time_t    rt      =   (*pfn)(entry);
    time_t          t;
#if defined(RECLS_PLATFORM_IS_WIN32)
    t = winstl__FILETIMEToUNIXTime(&rt, NULL);
#elif defined(RECLS_PLATFORM_IS_UNIX)
    t = rt;
#else
# error Platform not (yet) recognised
#endif /* platform */

    return rb_time_new(t, 0);
}

/** Creates a file entry instance, and adds in some of the more common fields as an optimisation. */
static VALUE Entry_create_(recls_info_t entry)
{
    VALUE   self = Data_Wrap_Struct(cEntry, NULL, Entry_free, (void*)entry);

    rb_iv_set(self, "@path", rb_str_new2(entry->path.begin));
#if defined(RECLS_PLATFORM_IS_WIN32)
    rb_iv_set(self, "@drive", rb_str_new(&entry->drive, 1));
#else /* RECLS_PLATFORM_IS_WIN32 */
    rb_iv_set(self, "@drive", rb_str_new("", 0));
#endif /* RECLS_PLATFORM_IS_WIN32 */
    rb_iv_set(self, "@file", rb_str_from_ptrs(entry->fileName.begin, entry->fileExt.end));

    return self;
}

static VALUE Entry_searchDirectory_get(VALUE self)
{
    recls_info_t    entry   =   Entry__entry_handle_get_(self);
    char            buffer[1001];
    size_t          cch     =   Recls_GetSearchDirectoryProperty(entry, buffer, RECLS_NUM_ELEMENTS(buffer));

    return rb_str_new(buffer, cch);
}

static VALUE Entry_searchRelativePath_get(VALUE self)
{
    recls_info_t    entry   =   Entry__entry_handle_get_(self);
    char            buffer[1001];
    size_t          cch     =   Recls_GetSearchRelativePathProperty(entry, buffer, RECLS_NUM_ELEMENTS(buffer));

    return rb_str_new(buffer, cch);
}

static VALUE Entry_directory_get(VALUE self)
{
    return Entry_strptr_get_(self, offsetof(struct recls_fileinfo_t, directory));
}

#ifdef WIN32
static VALUE Entry_shortFile_get(VALUE self)
{
    return Entry_strptr_get_(self, offsetof(struct recls_fileinfo_t, shortFile));
}
#endif /* ? WIN32 */

static VALUE Entry_fileName_get(VALUE self)
{
    return Entry_strptr_get_(self, offsetof(struct recls_fileinfo_t, fileName));
}

static VALUE Entry_fileExt_get(VALUE self)
{
    return Entry_strptr_get_(self, offsetof(struct recls_fileinfo_t, fileExt));
}

static VALUE Entry_directoryPath_get(VALUE self)
{
    recls_info_t    entry   =   Entry__entry_handle_get_(self);
    char            buffer[1001];
    size_t          cch     =   Recls_GetDirectoryPathProperty(entry, buffer, RECLS_NUM_ELEMENTS(buffer));

    return rb_str_new(buffer, cch);
}

static VALUE Entry_uncDrive_get(VALUE self)
{
    recls_info_t    entry   =   Entry__entry_handle_get_(self);
    char            buffer[1001];
    size_t          cch     =   Recls_GetUNCDriveProperty(entry, buffer, RECLS_NUM_ELEMENTS(buffer));

    return rb_str_new(buffer, cch);
}

/** Get the directory parts. */
static VALUE Entry_directoryParts_get(VALUE self)
{
    /* This is too simple. Just enumerate each directory part and push
     * it onto an array. Return the array to the caller.
     */

    recls_info_t                    entry   =   Entry__entry_handle_get_(self);
    VALUE                           parts   =   rb_ary_new();
    struct recls_strptrs_t const    *begin  =   entry->directoryParts.begin;
    struct recls_strptrs_t const    *end    =   entry->directoryParts.end;

    for(; begin != end; ++begin)
    {
        rb_ary_push(parts, rb_str_from_strptrs(begin));
    }

    return parts;
}

static VALUE Entry_creationTime_get(VALUE self)
{
    return Entry_time_get_(self, Recls_GetCreationTime);
}

static VALUE Entry_modificationTime_get(VALUE self)
{
    return Entry_time_get_(self, Recls_GetModificationTime);
}

static VALUE Entry_lastAccessTime_get(VALUE self)
{
    return Entry_time_get_(self, Recls_GetLastAccessTime);
}

static VALUE Entry_lastStatusChangeTime_get(VALUE self)
{
    return Entry_time_get_(self, Recls_GetLastStatusChangeTime);
}

static VALUE Entry_size_get(VALUE self)
{
    recls_info_t    entry   =   Entry__entry_handle_get_(self);

#ifdef HAVE_LONG_LONG
# ifdef RECLS_PLATFORM_IS_WIN32
    return rb_ull2inum(entry->size.QuadPart);
# else /* ? RECLS_PLATFORM_IS_WIN32 */
    return rb_ull2inum(entry->size);
# endif /* RECLS_PLATFORM_IS_WIN32 */
#else /* ? HAVE_LONG_LONG */
# ifdef RECLS_PLATFORM_IS_WIN32
    return rb_uint2big(entry->size.QuadPart);
# else /* ? RECLS_PLATFORM_IS_WIN32 */
    return rb_uint2big(entry->size);
# endif /* RECLS_PLATFORM_IS_WIN32 */
#endif /* HAVE_LONG_LONG */
}

static VALUE Entry_isReadOnly_get(VALUE self)
{
    return Recls_IsFileReadOnly(Entry__entry_handle_get_(self)) ? Qtrue : Qfalse;
}

static VALUE Entry_isDirectory_get(VALUE self)
{
    return Recls_IsFileDirectory(Entry__entry_handle_get_(self)) ? Qtrue : Qfalse;
}

static VALUE Entry_isFile_get(VALUE self)
{
    return Entry_isDirectory_get(self) ? Qfalse : Qtrue;
}

static VALUE Entry_isLink_get(VALUE self)
{
    return Recls_IsFileLink(Entry__entry_handle_get_(self)) ? Qtrue : Qfalse;
}

static VALUE Entry_exist_get(VALUE self)
{
    return Recls_DoesEntryExist(Entry__entry_handle_get_(self)) ? Qtrue : Qfalse;
}

static VALUE Entry_isUNC_get(VALUE self)
{
    return Recls_IsFileUNC(Entry__entry_handle_get_(self)) ? Qtrue : Qfalse;
}

/** Comparison function, used in ordering file entry values. */
static VALUE Entry_cmp(VALUE self, VALUE rhs)
{
    recls_info_t    entrySelf   =   Entry__entry_handle_get_(self);
    recls_info_t    entryRHS    =   Entry__entry_handle_get_(rhs);
#ifdef RECLS_PLATFORM_IS_WIN32
    int             iCmp        =   lstrcmpiA(entrySelf->path.begin, entryRHS->path.begin);
#else /* ? RECLS_PLATFORM_IS_WIN32 */
    int             iCmp        =   strcmp(entrySelf->path.begin, entryRHS->path.begin);
#endif /* RECLS_PLATFORM_IS_WIN32 */

    return iCmp < 0 ? INT2FIX(-1) : (0 < iCmp) ? INT2FIX(+1) : INT2FIX(0);
}

/* /////////////////////////////////////////////////////////////////////////////
 * FileSearch / FtpSearch
 */

static void Search_each_helper(hrecls_t hSrch, recls_rc_t rc)
{
    if(RECLS_SUCCEEDED(rc))
    {
        recls_info_t    entry;

        TRACETODEBUGGER("Search_each_helper() - processing entries\n");

        rc = Recls_GetDetails(hSrch, &entry);

        for(; RECLS_SUCCEEDED(rc); rc = Recls_GetNextDetails(hSrch, &entry))
        {
            TRACETODEBUGGER("Entry: %s\n", entry->path.begin);

            /* Yield, with the requisite entry */
            rb_yield(Entry_create_(entry));
        }

        Recls_SearchClose(hSrch);
    }

    /* Failure with no-more-data is ok. All others require that we raise an error. */
    if(RECLS_RC_NO_MORE_DATA != rc)
    {
        char    szError[501];
        char    szMessage[601];

        szError[Recls_GetErrorString(rc, szError, RECLS_NUM_ELEMENTS(szError))] = '\0';
        sprintf(szMessage, "Search failed: %s", szError);

        rb_raise(eReclsException, szMessage);
    }
}

/* /////////////////////////////////////////////////////////////////////////////
 * FileSearch
 */

#if 0
class FileSearch
  def initialize(searchRoot, pattern, flags)
    @searchRoot = searchRoot
    @pattern    = pattern
    @flags      = flags
  end
end
{}
#endif /* 0 */

/** Creates a FileSearch instance, and adds in the three member variables. */
static VALUE FileSearch_initialize(VALUE self, VALUE searchRoot, VALUE pattern, VALUE flags)
{
    /* Allow for the fact that searchRoot might be a FileEntry.
     */
    int         tSearchRoot =   rb_type(searchRoot);
    int         tPattern    =   rb_type(pattern);
    const int   tFlags      =   rb_type(flags);

    TRACETODEBUGGER("FileSearch_initialize(types = 0x%0x, 0x%0x, 0x%0x)\n", tSearchRoot, tPattern, tFlags);

    if(T_NIL == tSearchRoot)
    {
        searchRoot = s_currentDirectory;
    }
    else if((   T_OBJECT == tSearchRoot ||
                T_DATA == tSearchRoot) &&
            cEntry == rb_class_of(searchRoot))
    {
        ID      path;

        TRACETODEBUGGER("FileSearch_initialize(): searchRoot is an Entry instance.\n");

        path        =   rb_intern("path");
        searchRoot  =   rb_funcall(searchRoot, path, 0);
        tSearchRoot =   rb_type(searchRoot);
    }

    if(T_NIL == tPattern)
    {
        pattern = s_wildcardsAll;
    }
    else if((   T_OBJECT == tPattern ||
                T_DATA == tPattern) &&
            cEntry == rb_class_of(pattern))
    {
        ID      path;

        TRACETODEBUGGER("FileSearch_initialize(): pattern is an Entry instance.\n");

        path        =   rb_intern("path");
        pattern     =   rb_funcall(pattern, path, 0);
        tPattern    =   rb_type(pattern);
    }

    if( T_STRING == tSearchRoot &&
        T_STRING == tPattern &&
        T_FIXNUM == tFlags)
    {
        TRACETODEBUGGER("FileSearch_initialize(%s, %s, 0x%08x)\n", StringValuePtr(searchRoot), StringValuePtr(pattern), FIX2INT(flags));
    }
    else
    {
        TRACETODEBUGGER("FileSearch_initialize()\n");
    }

    Check_Type(searchRoot, T_STRING);
    Check_Type(pattern, T_STRING);
    Check_Type(flags, T_FIXNUM);

    // The FileSearch class has three members: searchRoot, pattern, flags
    rb_iv_set(self, "@searchRoot", searchRoot);
    rb_iv_set(self, "@pattern", pattern);
    rb_iv_set(self, "@flags", flags);

    return self;
}

/** The each() method. */

static VALUE FileSearch_each(VALUE self)
{
    VALUE           searchRoot  =   rb_iv_get(self, "@searchRoot");
    VALUE           pattern     =   rb_iv_get(self, "@pattern");
    recls_uint32_t  flags       =   FIX2INT(rb_iv_get(self, "@flags"));

    hrecls_t        hSrch;
    recls_rc_t      rc;

    rc = Recls_Search(StringValuePtr(searchRoot), StringValuePtr(pattern), flags, &hSrch);

    Search_each_helper(hSrch, rc);

    return self;
}

static VALUE FileSearch_to_s(VALUE self)
{
    VALUE           searchRoot_ =   rb_iv_get(self, "@searchRoot");
    VALUE           pattern_    =   rb_iv_get(self, "@pattern");
    VALUE           flags_      =   rb_iv_get(self, "@flags");
    char  const     *searchRoot =   StringValuePtr(searchRoot_);
    char  const     *pattern    =   StringValuePtr(pattern_);
    recls_uint32_t  flags       =   FIX2INT(flags_);
    size_t          cch         =   100 + strlen(searchRoot) + strlen(pattern);
    char            *sz         =   (char*)alloca(1 + cch);

#ifdef _MSC_VER
    cch = _snprintf(sz
                ,   cch
#else /* ? compiler */
    cch = sprintf(  sz
#endif /* compiler */
                ,   "Recls::FileSearch:0x%08x (searchRoot: %s; pattern: %s; flags: 0x%08x)"
                ,   self
                ,   searchRoot
                ,   pattern
                ,   flags);

    return rb_str_new(sz, cch);
}

/* /////////////////////////////////////////////////////////////////////////////
 * FtpSearch
 */

#if 0
class FtpSearch
  def initialize(searchRoot, pattern, flags)
    @searchRoot = searchRoot
    @pattern    = pattern
    @flags      = flags
  end
end
{}
#endif /* 0 */

/** Creates a FtpSearch instance, and adds in the three member variables. */
static VALUE FtpSearch_initialize(VALUE self, VALUE host, VALUE username, VALUE password, VALUE searchRoot, VALUE pattern, VALUE flags)
{
    TRACETODEBUGGER("FtpSearch_initialize(): searchRoot: %s\n", StringValuePtr(searchRoot));

    Check_Type(host, T_STRING);
    Check_Type(username, T_STRING);
    Check_Type(password, T_STRING);
    Check_Type(searchRoot, T_STRING);
    Check_Type(pattern, T_STRING);
    Check_Type(flags, T_FIXNUM);

    // The FtpSearch class has three members: searchRoot, pattern, flags
    rb_iv_set(self, "@host", host);
    rb_iv_set(self, "@username", username);
    rb_iv_set(self, "@password", password);
    rb_iv_set(self, "@searchRoot", searchRoot);
    rb_iv_set(self, "@pattern", pattern);
    rb_iv_set(self, "@flags", flags);

    return self;
}

/** The each() method. */
static VALUE FtpSearch_each(VALUE self)
{
    VALUE           host        =   rb_iv_get(self, "@host");
    VALUE           username    =   rb_iv_get(self, "@username");
    VALUE           password    =   rb_iv_get(self, "@password");
    VALUE           searchRoot  =   rb_iv_get(self, "@searchRoot");
    VALUE           pattern     =   rb_iv_get(self, "@pattern");
    recls_uint32_t  flags       =   FIX2INT(rb_iv_get(self, "@flags"));

    hrecls_t        hSrch;
    recls_rc_t      rc;

    rc = Recls_SearchFtp(StringValuePtr(host), StringValuePtr(username), StringValuePtr(password), StringValuePtr(searchRoot), StringValuePtr(pattern), flags, &hSrch);

    Search_each_helper(hSrch, rc);

    return self;
}

static VALUE FtpSearch_to_s(VALUE self)
{
    VALUE           host_       =   rb_iv_get(self, "@host");
    VALUE           username_   =   rb_iv_get(self, "@username");
/*     VALUE           password_    =   rb_iv_get(self, "@password"); */
    VALUE           searchRoot_ =   rb_iv_get(self, "@searchRoot");
    VALUE           pattern_    =   rb_iv_get(self, "@pattern");
    char  const     *host       =   StringValuePtr(host_);
    char  const     *username   =   StringValuePtr(username_);
/*     char  const      *password   =   StringValuePtr(password_); */
    char  const     *searchRoot =   StringValuePtr(searchRoot_);
    char  const     *pattern    =   StringValuePtr(pattern_);
    recls_uint32_t  flags       =   FIX2INT(rb_iv_get(self, "@flags"));
    size_t          cch         =   100 + strlen(host) + strlen(username) + /* strlen(password) +  */strlen(searchRoot) + strlen(pattern);
    char            *sz         =   (char*)alloca(1 + cch);

#ifdef _MSC_VER
    cch = _snprintf(sz
                ,   cch
#else /* ? compiler */
    cch = sprintf(  sz
#endif /* compiler */
                ,   "Recls::FtpSearch:0x%08x (host: %s; username: %s; password: %s; searchRoot: %s; pattern: %s; flags: 0x%08x)"
                ,   self
                ,   host
                ,   username
                ,   "****"
                ,   searchRoot
                ,   pattern
                ,   flags);

    return rb_str_new(sz, cch);
}

/* /////////////////////////////////////////////////////////////////////////////
 * Module functions
 */

static VALUE recls_showTrace_get(VALUE self)
{
    STLSOFT_SUPPRESS_UNUSED(self);

    return INT2FIX(0);
}

static VALUE recls_showTrace_set(VALUE self, VALUE value)
{
    STLSOFT_SUPPRESS_UNUSED(self);
    STLSOFT_SUPPRESS_UNUSED(value);

    return Qnil;
}

/** Returns an array of roots for the current system's file-system. */
static VALUE recls_roots_get(VALUE self)
{
    VALUE           parts   =   rb_ary_new();
    recls_root_t    roots[26];
    size_t          cRoots  =   Recls_GetRoots(roots, RECLS_NUM_ELEMENTS(roots));
    size_t          i;

    STLSOFT_SUPPRESS_UNUSED(self);

    for(i = 0; i < cRoots; ++i)
    {
        rb_ary_push(parts, rb_str_new2(roots[i].name));
    }

    return parts;
}

static VALUE recls_selectedRoots_get_(VALUE self, recls_uint32_t flags)
{
    VALUE           parts   =   rb_ary_new();
    recls_root_t    roots[26];
    size_t          cRoots  =   Recls_GetSelectedRoots(roots, RECLS_NUM_ELEMENTS(roots), flags);
    size_t          i;

    /* printf("recls_selectedRoots_get_(, 0x%08x)\n", flags); */

    STLSOFT_SUPPRESS_UNUSED(self);

    for(i = 0; i < cRoots; ++i)
    {
        rb_ary_push(parts, rb_str_new2(roots[i].name));
    }

    return parts;
}

static VALUE recls_selectedRoots_get(VALUE self, VALUE flags)
{
//    Check_Type(flags, T_FIXNUM);

    /* printf("recls_selectedRoots_get(, 0x%08x)\n", flags); */

    STLSOFT_SUPPRESS_UNUSED(self);

    return recls_selectedRoots_get_(self, FIX2INT(flags));
}

/** Retrieves the path name separator symbol, e.g. "/". */
static VALUE recls_pathNameSeparator_get(VALUE self)
{
    STLSOFT_SUPPRESS_UNUSED(self);

    return rb_str_new2(Recls_GetPathNameSeparator());
}

/** Retrieves the path separator symbol, e.g. ";". */
static VALUE recls_pathSeparator_get(VALUE self)
{
    STLSOFT_SUPPRESS_UNUSED(self);

    return rb_str_new2(Recls_GetPathSeparator());
}

/** Retrieves the wildcards-all symbol, e.g. "*". */
static VALUE recls_wildcardsAll_get(VALUE self)
{
    STLSOFT_SUPPRESS_UNUSED(self);

    return rb_str_new2(Recls_GetWildcardsAll());
}

/** Retrieves the file entry matching the given  */
static VALUE recls_stat_(char const *filePath, recls_uint32_t flags)
{
    VALUE           ret     =   Qnil;
    recls_info_t    entry;
    recls_rc_t      rc      =   Recls_Stat(filePath, flags, &entry);

    if(RECLS_SUCCEEDED(rc))
    {
        TRACETODEBUGGER("Entry: %s\n", entry->path.begin);

        ret = Entry_create_(entry);
    }

    /* Failure with no-more-data is ok. All others require that we raise an error. */
    if( RECLS_FAILED(rc) &&
        RECLS_RC_NO_MORE_DATA != rc)
    {
        char    szError[501];
        char    szMessage[601];

        szError[Recls_GetErrorString(rc, szError, RECLS_NUM_ELEMENTS(szError))] = '\0';
        sprintf(szMessage, "Failed to stat '%s': %s", filePath, szError);

        rb_raise(eReclsException, szMessage);
    }

    return ret;
}

static VALUE recls_stat_1(VALUE self, VALUE pattern)
{
    STLSOFT_SUPPRESS_UNUSED(self);

    Check_Type(pattern, T_STRING);

    return recls_stat_(StringValuePtr(pattern), RECLS_F_FILES | RECLS_F_DIRECTORIES);
}

static VALUE recls_stat_2(VALUE self, VALUE pattern, VALUE flags)
{
    STLSOFT_SUPPRESS_UNUSED(self);

    Check_Type(pattern, T_STRING);
    Check_Type(flags, T_FIXNUM);

    return recls_stat_(StringValuePtr(pattern), FIX2INT(flags));
}

static VALUE recls_stat_v(int argc, VALUE *argv, VALUE self)
{
    VALUE   ret;

    switch(argc)
    {
        case    1:
            ret = recls_stat_1(self, argv[0]);
            break;
        case    2:
            ret = recls_stat_2(self, argv[0], argv[1]);
            break;
        default:
            rb_raise(rb_eArgError, "wrong number of arguments (%d)", argc);
            break;
    }

    return ret;
}

/* /////////////////////////////////////////////////////////////////////////////
 * module functions
 */

static VALUE recls_usage(VALUE self)
{
    VALUE       ar          =   rb_ary_new();
    char const  *lines[]    =
    {
            "Module: Recls, version " \
                __synsoft_stringize_symbol(RECLS_VER_C_RECLS_RUBY_MAJOR) "." \
                __synsoft_stringize_symbol(RECLS_VER_C_RECLS_RUBY_MINOR) "." \
                __synsoft_stringize_symbol(RECLS_VER_C_RECLS_RUBY_REVISION) "." \
                __synsoft_stringize_symbol(RECLS_VER_C_RECLS_RUBY_EDIT) "."
        ,   ""
        ,   "Author: Matthew Wilson, Synesis Software; http://synesis.com.au/software/"
        ,   ""
        ,   "  Constants:"
        ,   "    FILES                                - files are included in the search"
        ,   "    DIRECTORIES                          - directories are included in the search"
        ,   "    LINKS                                - links are included in the search"
        ,   "    DEVICES                              - devices are included in the search"
        ,   "    TYPEMASK                             - mask for types"
        ,   "    RECURSIVE                            - search is recursive"
        ,   "    NO_FOLLOW_LINKS                      - links are not followed"
        ,   "    DIRECTORY_PARTS                      - directory parts are evaluated"
        ,   "    DETAILS_LATER                        - details are fetched later (not currently"
        ,   "                                           supported)"
        ,   ""
        ,   "    RECLS_VERSION                        - version of the recls library"
        ,   "    VERSION                              - version of the recls/Ruby mapping"
        ,   ""
        ,   "    FIXED_DRIVES                         - causes selectedRoots() to include local"
        ,   "                                           fixed drives"
        ,   "    NETWORK_DRIVES                       - causes selectedRoots() to include network"
        ,   "                                           drives"
        ,   "    CDROM_DRIVES                         - causes selectedRoots() to include local"
        ,   "                                           optical drives"
        ,   "    REMOVABLE_DRIVES                     - causes selectedRoots() to include local"
        ,   "                                           removable drives"
        ,   "    RAM_DRIVES                           - causes selectedRoots() to include memory"
        ,   "                                           drives"
        ,   ""
        ,   "  Attributes:"
        ,   "    roots                                - returns an array of file system roots (e.g."
        ,   "                                           { \"/\"} for UNIX,"
        ,   "                                           { \"C:\\\", \"H:\\\", ... } for Windows)"
        ,   "    pathNameSeparator                    - returns the path-name separator (e.g."
        ,   "                                           \"/\" for UNIX, \"\\\" for Windows)"
        ,   "    wildcardsAll                         - returns the \"all\" wildcards string (e.g."
        ,   "                                           \"*\" for UNIX, \"*.*\" for Windows)"
        ,   ""
        ,   "  Functions:"
        ,   "    usage                                - shows this usage"
        ,   ""
        ,   "    stat (path [, flags])                - retrieves the Entry information for the"
        ,   "                                           given path/flags"
        ,   "    selectedRoots (flags)                - like roots, but takes a combination of"
        ,   "                                           flags that determine which file system"
        ,   "                                           roots are included"
        ,   "    combinePaths (path1, path2)          - combines two path fragments. If path1 is"
        ,   "                                           an Entry instance, then it is converted to"
        ,   "                                           a string via the 'path' attribute. If path2"
        ,   "                                           is an Entry instance, then it is converted"
        ,   "                                           to a string via the 'searchRelativePath'"
        ,   "                                           attribute."
        ,   "    deriveRelativePath (origin, target)  - determines the relative path between an"
        ,   "                                           origin and a target path"
        ,   "    squeezePath (path, width)            - writes the path into the given width"
        ,   ""
        ,   "  Classes:"
        ,   "    Class: FileSearch"
        ,   "      Construction:"
        ,   "        new (searchRoot, pattern, flags)"
        ,   "          searchRoot                     - the root directory of the search"
        ,   "          pattern                        - the search pattern (e.g. \"*.rb;stlsoft*.h\")"
        ,   "          flags                          - flags to alter the behaviour of the search"
        ,   "      Attributes:"
        ,   "        searchRoot                       - the root directory specified in the"
        ,   "                                            constructor"
        ,   "        pattern                          - the pattern specified in the constructor"
        ,   "        flags                            - the flags specified in the constructor"
        ,   "      Methods:"
        ,   "        each                             - enumerates the items (of type Entry) located"
        ,   "                                           by the search"
        ,   "        to_s                             - returns a string form of the search object"
        ,   "    Class: FtpSearch"
        ,   "      Construction:"
        ,   "        new (host, username, password, searchRoot, pattern, flags)"
        ,   "          host                           - the host on which to search"
        ,   "          username                       - the username of the account on the host"
        ,   "          password                       - the password of the account on the host"
        ,   "          searchRoot                     - the root directory of the search"
        ,   "          pattern                        - the search pattern (e.g. \"*.rb;stlsoft*.h\")"
        ,   "          flags                          - flags to alter the behaviour of the search"
        ,   "      Attributes:"
        ,   "        host                             - the host name specified in the constructor"
        ,   "        username                         - the username specified in the constructor"
        ,   "        searchRoot                       - the root directory specified in the"
        ,   "                                           constructor"
        ,   "        pattern                          - the pattern specified in the constructor"
        ,   "        flags                            - the flags specified in the constructor"
        ,   "      Methods:"
        ,   "        each                             - enumerates the items (of type Entry) located"
        ,   "                                           by the search"
        ,   "        to_s                             - returns a string form of the search object"
        ,   "    Class: Entry"
        ,   "      Attributes:"
        ,   "        path                             - the full path of the entry"
        ,   "        searchDirectory                  - the directory specified to the FileSearch"
        ,   "                                           constructor"
        ,   "        searchRelativePath               - the path of the entry, relative to the search"
        ,   "                                           directory"
        ,   "        drive                            - the drive part of the entry's path (Windows"
        ,   "                                           only)"
        ,   "        directory                        - the directory part of the entry's path"
        ,   "        directoryPath                    - the directory + drive part of the entry's path"
        ,   "        uncDrive                         - the UNC part of the entry's path (Windows"
        ,   "                                           only)"
        ,   "        directoryParts                   - an array of the individual parts of the"
        ,   "                                           entry's directory"
        ,   "        file                             - the entry's file name (including extension)"
        ,   "        fileName                         - the entry's file name (excluding extension)"
        ,   "        fileExt                          - the entry's extension"
        ,   "        shortFile                        - the short form of the file name (this is the"
        ,   "                                           same as file on non-Windows systems)"
        ,   "        creationTime                     - the time the entry was created"
        ,   "        modificationTime                 - the time the entry was modified"
        ,   "        lastAccessTime                   - the time the entry was last accessed"
        ,   "        lastStatusChangeTime             - the time the entry's status was last changed"
        ,   "        size                             - the size of the entry (0 for directories)"
        ,   "        isReadOnly                       - indicates whether the entry is read-only"
        ,   "        isDirectory                      - indicates whether the entry is a directory"
        ,   "        isLink                           - indicates whether the entry is a link"
        ,   "        isUNC                            - indicates whether the entry's path is UNC"
        ,   "        to_s                             - returns a string form of the entry object"
    };
    char const  **begin =   &lines[0];
    char const  **end   =   &lines[RECLS_NUM_ELEMENTS(lines)];

    for(; begin != end; ++begin)
    {
        TRACETODEBUGGER("%s\n", *begin);
        rb_ary_push(ar, rb_str_new2(*begin));
    }

    STLSOFT_SUPPRESS_UNUSED(self);

    return ar;
}


static VALUE recls_combinePaths_(char const *path1, char const *path2)
{
    char    *result;
    size_t  cchResult;

    TRACETODEBUGGER("combinePaths(%s, %s)\n", (NULL == path1) ? "" : path1, (NULL == path2) ? "" : path2);

    cchResult   =   Recls_CombinePaths(path1, path2, NULL, 0);
    result      =   (char*)alloca(1 + cchResult);

    Recls_CombinePaths(path1, path2, result, cchResult);

    return rb_str_new(result, cchResult);
}

static VALUE recls_deriveRelativePath_(char const *origin, char const *target)
{
    char    *result;
    size_t  cchResult;

    TRACETODEBUGGER("deriveRelativePath(%s, %s)\n", origin, target);

    cchResult   =   Recls_DeriveRelativePath(origin, target, NULL, 0);
    result      =   (char*)alloca(1 + cchResult);

    Recls_DeriveRelativePath(origin, target, result, cchResult);

    return rb_str_new(result, cchResult);
}

static VALUE recls_squeezePath_(char const *path, unsigned width)
{
    char    *result;

    TRACETODEBUGGER("squeezePath(%s, %u)\n", path, width);

    result      =   (char*)alloca(width);

    Recls_SqueezePath(path, result, width);

    return rb_str_new(result, width);
}


static VALUE recls_combinePaths(VALUE self, VALUE path1, VALUE path2)
{
    STLSOFT_SUPPRESS_UNUSED(self);

    if( NIL_P(path1) &&
        NIL_P(path2))
    {
        rb_raise(rb_eArgError, "at most one argument can be nil");
    }
    else
    {
        int         tPath1  =   rb_type(path1);
        int         tPath2  =   rb_type(path2);
        char const  *strPath1;
        char const  *strPath2;

        if(T_NIL == tPath1)
        {
            strPath1 = NULL;
        }
        else
        {
            if( (   T_OBJECT == tPath1 ||
                    T_DATA == tPath1) &&
                cEntry == rb_class_of(path1))
            {
                ID  method  =   rb_intern("path");

                path1       =   rb_funcall(path1, method, 0);
                tPath1      =   rb_type(path1);
            }

            Check_Type(path1, T_STRING);

            strPath1 = StringValuePtr(path1);
        }

        if(T_NIL == tPath2)
        {
            strPath2 = NULL;
        }
        else
        {
            if( (   T_OBJECT == tPath2 ||
                    T_DATA == tPath2) &&
                cEntry == rb_class_of(path2))
            {
                ID  method  =   rb_intern("searchRelativePath");

                path2       =   rb_funcall(path2, method, 0);
                tPath2      =   rb_type(path2);
            }

            Check_Type(path2, T_STRING);

            strPath2 = StringValuePtr(path2);
        }

        TRACETODEBUGGER("recls_combinePaths(%s, %s)\n", strPath1, strPath2);

        return recls_combinePaths_(strPath1, strPath2);
    }
}

static VALUE recls_deriveRelativePath(VALUE self, VALUE origin, VALUE target)
{
    Check_Type(origin, T_STRING);
    Check_Type(target, T_STRING);

    STLSOFT_SUPPRESS_UNUSED(self);

    return recls_deriveRelativePath_(StringValuePtr(origin), StringValuePtr(target));
}

static VALUE recls_squeezePath(VALUE self, VALUE path, VALUE width)
{
    Check_Type(path, T_STRING);

    STLSOFT_SUPPRESS_UNUSED(self);

    return recls_squeezePath_(StringValuePtr(path), FIX2UINT(width));
}

/* /////////////////////////////////////////////////////////////////////////////
 * Module registration
 */

void Init_recls()
{
    char    reclsVersion[101];
    char    reclsRubyVersion[101];

    TRACETODEBUGGER("Initialising recls extension for Ruby\n");

    sprintf(reclsVersion, "%d.%d.%d", RECLS_VER_MAJOR, RECLS_VER_MINOR, RECLS_VER_REVISION);
    sprintf(reclsRubyVersion, "%d.%d.%d", RECLS_VER_C_RECLS_RUBY_MAJOR, RECLS_VER_C_RECLS_RUBY_MINOR, RECLS_VER_C_RECLS_RUBY_REVISION);

    /* Global variables */
    s_currentDirectory  =   rb_str_new(".", 1);
    s_wildcardsAll      =   rb_str_new2(Recls_GetWildcardsAll());

    /* Recls module */
    mRecls = rb_define_module("Recls");
    rb_define_const(mRecls, "RECLS_VERSION",rb_str_new2(reclsVersion));
    rb_define_const(mRecls, "VERSION",rb_str_new2(reclsRubyVersion));
    rb_define_const(mRecls, "RECLS_F_FILES", rb_uint2inum(RECLS_F_FILES));
    rb_define_const(mRecls, "RECLS_F_DIRECTORIES", rb_uint2inum(RECLS_F_DIRECTORIES));
    rb_define_const(mRecls, "RECLS_F_LINKS", rb_uint2inum(RECLS_F_LINKS));
    rb_define_const(mRecls, "RECLS_F_DEVICES", rb_uint2inum(RECLS_F_DEVICES));
    rb_define_const(mRecls, "RECLS_F_TYPEMASK", rb_uint2inum(RECLS_F_TYPEMASK));
    rb_define_const(mRecls, "RECLS_F_RECURSIVE", rb_uint2inum(RECLS_F_RECURSIVE));
    rb_define_const(mRecls, "RECLS_F_NO_FOLLOW_LINKS", rb_uint2inum(RECLS_F_NO_FOLLOW_LINKS));
    rb_define_const(mRecls, "RECLS_F_DIRECTORY_PARTS", rb_uint2inum(RECLS_F_DIRECTORY_PARTS));
    rb_define_const(mRecls, "RECLS_F_DETAILS_LATER", rb_uint2inum(RECLS_F_DETAILS_LATER));
    rb_define_const(mRecls, "FILES", rb_uint2inum(RECLS_F_FILES));
    rb_define_const(mRecls, "DIRECTORIES", rb_uint2inum(RECLS_F_DIRECTORIES));
    rb_define_const(mRecls, "LINKS", rb_uint2inum(RECLS_F_LINKS));
    rb_define_const(mRecls, "DEVICES", rb_uint2inum(RECLS_F_DEVICES));
    rb_define_const(mRecls, "TYPEMASK", rb_uint2inum(RECLS_F_TYPEMASK));
    rb_define_const(mRecls, "RECURSIVE", rb_uint2inum(RECLS_F_RECURSIVE));
    rb_define_const(mRecls, "NO_FOLLOW_LINKS", rb_uint2inum(RECLS_F_NO_FOLLOW_LINKS));
    rb_define_const(mRecls, "DIRECTORY_PARTS", rb_uint2inum(RECLS_F_DIRECTORY_PARTS));
    rb_define_const(mRecls, "DETAILS_LATER", rb_uint2inum(RECLS_F_DETAILS_LATER));
    rb_define_module_function(mRecls, "showTrace", recls_showTrace_get, 0);
    rb_define_module_function(mRecls, "showTrace=", recls_showTrace_set, 1);
    rb_define_module_function(mRecls, "roots", recls_roots_get, 0);
    rb_define_module_function(mRecls, "pathNameSeparator", recls_pathNameSeparator_get, 0);
    rb_define_module_function(mRecls, "pathSeparator", recls_pathSeparator_get, 0);
    rb_define_module_function(mRecls, "wildcardsAll", recls_wildcardsAll_get, 0);
    rb_define_module_function(mRecls, "stat", recls_stat_v, -1);
    rb_define_module_function(mRecls, "usage", recls_usage, 0);

    rb_define_module_function(mRecls, "combinePaths", recls_combinePaths, 2);
    rb_define_module_function(mRecls, "deriveRelativePath", recls_deriveRelativePath, 2);
    rb_define_module_function(mRecls, "squeezePath", recls_squeezePath, 2);

    rb_define_const(mRecls, "FIXED_DRIVES", rb_uint2inum(RECLS_F_FIXED_DRIVES));
    rb_define_const(mRecls, "NETWORK_DRIVES", rb_uint2inum(RECLS_F_NETWORK_DRIVES));
    rb_define_const(mRecls, "CDROM_DRIVES", rb_uint2inum(RECLS_F_CDROM_DRIVES));
    rb_define_const(mRecls, "REMOVABLE_DRIVES", rb_uint2inum(RECLS_F_REMOVABLE_DRIVES));
    rb_define_const(mRecls, "RAM_DRIVES", rb_uint2inum(RECLS_F_RAM_DRIVES));
    rb_define_module_function(mRecls, "selectedRoots", recls_selectedRoots_get, 1);

    /* Recls exception */
    eReclsException = rb_define_class_under(mRecls, "ReclsException", rb_eRuntimeError);

    /* FileSearch class */
    cFileSearch = rb_define_class_under(mRecls, "FileSearch", rb_cObject);
    rb_include_module(cFileSearch, rb_mEnumerable);
    rb_define_method(cFileSearch, "initialize", FileSearch_initialize, 3);
    rb_define_method(cFileSearch, "each", FileSearch_each, 0);
    rb_define_method(cFileSearch, "to_s", FileSearch_to_s, 0);
    rb_define_attr(cFileSearch,   "searchRoot", 1, 0);
    rb_define_attr(cFileSearch,   "pattern", 1, 0);
    rb_define_attr(cFileSearch,   "flags", 1, 0);

    /* FtpSearch class */
    cFtpSearch = rb_define_class_under(mRecls, "FtpSearch", rb_cObject);
    rb_include_module(cFtpSearch, rb_mEnumerable);
    rb_define_method(cFtpSearch, "initialize", FtpSearch_initialize, 6);
    rb_define_method(cFtpSearch, "each", FtpSearch_each, 0);
    rb_define_method(cFtpSearch, "to_s", FtpSearch_to_s, 0);
    rb_define_attr(cFtpSearch,   "host", 1, 0);
    rb_define_attr(cFtpSearch,   "username", 1, 0);
/*     rb_define_attr(cFtpSearch,   "password", 1, 0); */
    rb_define_attr(cFtpSearch,   "searchRoot", 1, 0);
    rb_define_attr(cFtpSearch,   "pattern", 1, 0);
    rb_define_attr(cFtpSearch,   "flags", 1, 0);

    /* Entry class */
    cEntry = rb_define_class_under(mRecls, "Entry", rb_cObject);
    rb_include_module(cEntry, rb_mComparable);
    rb_define_attr(cEntry,   "path", 1, 0);
    rb_define_method(cEntry, "searchDirectory", Entry_searchDirectory_get, 0);
    rb_define_method(cEntry, "searchRelativePath", Entry_searchRelativePath_get, 0);
    rb_define_attr(cEntry,   "drive", 1, 0);
    rb_define_method(cEntry, "directory", Entry_directory_get, 0);
    rb_define_method(cEntry, "directoryPath", Entry_directoryPath_get, 0);
    rb_define_method(cEntry, "uncDrive", Entry_uncDrive_get, 0);
    rb_define_method(cEntry, "directoryParts", Entry_directoryParts_get, 0);
    rb_define_attr(cEntry,   "file", 1, 0);
#ifdef WIN32
    rb_define_method(cEntry, "shortFile", Entry_shortFile_get, 0);
#else /* ? WIN32 */
    rb_define_alias(cEntry,  "shortFile", "file");
#endif /* ? WIN32 */
    rb_define_method(cEntry, "fileName", Entry_fileName_get, 0);
    rb_define_method(cEntry, "fileExt", Entry_fileExt_get, 0);
    rb_define_method(cEntry, "creationTime", Entry_creationTime_get, 0);
    rb_define_method(cEntry, "modificationTime", Entry_modificationTime_get, 0);
    rb_define_method(cEntry, "lastAccessTime", Entry_lastAccessTime_get, 0);
    rb_define_method(cEntry, "lastStatusChangeTime", Entry_lastStatusChangeTime_get, 0);
    rb_define_method(cEntry, "size", Entry_size_get, 0);
    rb_define_method(cEntry, "isReadOnly", Entry_isReadOnly_get, 0);
    rb_define_method(cEntry, "isDirectory", Entry_isDirectory_get, 0);
    rb_define_method(cEntry, "isFile", Entry_isFile_get, 0);
    rb_define_method(cEntry, "isLink", Entry_isLink_get, 0);
    rb_define_method(cEntry, "exist?", Entry_exist_get, 0);
    rb_define_method(cEntry, "isUNC", Entry_isUNC_get, 0);
    rb_define_method(cEntry, "<=>", Entry_cmp, 1);
    rb_define_alias(cEntry,  "to_s", "path");
}

/* ////////////////////////////////////////////////////////////////////////// */
