/* /////////////////////////////////////////////////////////////////////////////
 * File:        ReclsFileSearch_win32.cpp
 *
 * Purpose:     Implementation of the ReclsFileSearch class for Win32.
 *
 * Created:     16th August 2003
 * Updated:     27th February 2007
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2007, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls/recls.h>
#include <recls/assert.h>
#include "recls_string.h"
#include "recls_util.h"
#include "recls_impl.h"

#if RECLS_DEBUG_LEVEL > 0
# include <ctype.h>
#endif /* RECLS_DEBUG_LEVEL > 0 */

#include "recls_debug.h"

#include "ReclsSearch.hpp"
#include "ReclsFileSearch.hpp"
#include "ReclsFileSearchDirectoryNode.hpp"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

namespace {

typedef ReclsFileSearchDirectoryNode::traits_type               traits_type;
typedef ReclsFileSearchDirectoryNode::file_path_buffer_type     file_path_buffer_type;

stlsoft_ns_using(c_str_ptr)
stlsoft_ns_using(c_str_len)
stlsoft_ns_using(sap_cast)

} // anonymous namespace

/* /////////////////////////////////////////////////////////////////////////////
 * Utility functions
 */

inline size_t count_dir_parts(recls_char_t const *begin, recls_char_t const *end)
{
    return count_char_instances(begin, end, traits_type::path_name_separator());
}

/* /////////////////////////////////////////////////////////////////////////////
 * ReclsFileSearch
 */

inline void *ReclsFileSearch::operator new(size_t cb, size_t cDirParts, size_t cbRootDir)
{
    function_scope_trace("ReclsFileSearch::operator new");

    cbRootDir = align_up_size(cbRootDir);

#if !defined(STLSOFT_COMPILER_IS_GCC)
    RECLS_ASSERT(cb > STLSOFT_RAW_OFFSETOF(ReclsFileSearch, data));
#endif /* compiler */

    cb  =   STLSOFT_RAW_OFFSETOF(ReclsFileSearch, data);
    cb  +=  (cDirParts) * sizeof(recls_strptrs_t);
    cb  +=  cbRootDir;

    void    *pv =   malloc(cb);

#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
    if(NULL == pv)
    {
        throw std::bad_alloc();
    }
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */

    return pv;
}

#ifdef RECLS_COMPILER_REQUIRES_MATCHING_PLACEMENT_DELETE
inline void ReclsFileSearch::operator delete(void *pv, size_t /* cDirParts */, size_t /* cbRootDir */)
{
    function_scope_trace("ReclsFileSearch::operator delete");

    free(pv);
}
#endif /* RECLS_COMPILER_REQUIRES_MATCHING_PLACEMENT_DELETE */

inline void ReclsFileSearch::operator delete(void *pv)
{
    function_scope_trace("ReclsFileSearch::operator delete");

    free(pv);
}

/* static */ recls_rc_t ReclsFileSearch::FindAndCreate( recls_char_t const          *rootDir
                                                    ,   recls_char_t const          *pattern
                                                    ,   recls_uint32_t              flags
                                                    ,   hrecls_progress_fn_t        pfn
                                                    ,   recls_process_fn_param_t    param
                                                    ,   ReclsFileSearch             **ppsi)
{
	RECLS_ASSERT(NULL != rootDir);
	RECLS_ASSERT(NULL != pattern);

    function_scope_trace("ReclsFileSearch::FindAndCreate");

//  debug_printf("Searching: %s / %s / 0x%08x\n", rootDir, pattern, flags);

    recls_rc_t              rc;
    file_path_buffer_type   fullPath;
    size_t                  cchFullPath;

    *ppsi = NULL;

    cchFullPath = traits_type::get_full_path_name(rootDir, fullPath.size(), &fullPath[0]);
    if( 0 == cchFullPath ||
        (traits_type::remove_dir_end(&fullPath[0]), !traits_type::file_exists(c_str_ptr(fullPath))))
    {
        rc = RECLS_RC_INVALID_DIRECTORY;
    }
    else
    {
#if defined(RECLS_PLATFORM_IS_UNIX) && \
    defined(WIN32) && \
    defined(EMULATE_UNIX_ON_WIN32)
        // On Win32 systems that are being used to test the UNIX build, we will
        // translate all the backslashes in UNIX-like paths to forward slashes.
        {
            recls_char_t *const dir0    =   const_cast<recls_char_t *>(find_directory_0(&fullPath[0]));
            recls_char_t *const end     =   &fullPath[0] + cchFullPath;

            stlsoft_ns_qual_std(replace)(dir0, end, '\\', traits_type::path_name_separator());
        }

        RECLS_STRING_TEMPLATE_1(recls_char_t)   pattern_(pattern);

        stlsoft_ns_qual_std(replace)(pattern_.begin(), pattern_.end(), ';', ':');

        pattern = c_str_ptr(pattern_);

#endif /* RECLS_PLATFORM_IS_UNIX && WIN32 && EMULATE_UNIX_ON_WIN32 */

        traits_type::ensure_dir_end(&fullPath[0]);

        size_t      lenSearchRoot   =   c_str_len(fullPath);

        RECLS_ASSERT(0 < lenSearchRoot);

        rootDir = c_str_ptr(fullPath);

        // Count the directory parts. This is always done for the ReclsFileSearch class, since it
        // uses them to recurse.
        recls_char_t const          *dir0       =   find_directory_0(rootDir);
        recls_char_t const *const   end         =   rootDir + lenSearchRoot;
        const size_t                cDirParts   =   count_dir_parts(dir0, end);

        // Create the search handle
        ReclsFileSearch         *si;

#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
        try
        {
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */
            si = new(cDirParts, sizeof(recls_char_t) * (1 + lenSearchRoot)) ReclsFileSearch(cDirParts, rootDir, lenSearchRoot, pattern, pfn, param, flags);
#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
        }
        catch(std::bad_alloc &)
        {
            si = NULL;
        }
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */

        if(NULL == si)
        {
            rc = RECLS_RC_FAIL;
        }
        else
        {
            // This is a nasty hack. It's tantamount to ctor & create function, so
            // should be made more elegant soon.
            if(NULL == si->m_dnode)
            {
                delete si;

                rc = RECLS_RC_NO_MORE_DATA;
            }
            else
            {
                *ppsi = si;

                rc = RECLS_RC_OK;
            }
        }
    }

    RECLS_ASSERT(RECLS_RC_OK == rc || NULL == *ppsi);

    return rc;
}

recls_char_t const *ReclsFileSearch::calc_rootDir_(size_t cDirParts, recls_char_t const *rootDir)
{
    function_scope_trace("ReclsFileSearch::calc_rootDir_");

    // Root dir is located after file parts, and before pattern
    return traits_type::str_copy(::stlsoft::sap_cast<recls_char_t*>(&data[cDirParts * sizeof(recls_strptrs_t)]), rootDir);
}

ReclsFileSearch::ReclsFileSearch(   size_t                      cDirParts
                                ,   recls_char_t const          *rootDir
                                ,   size_t                      rootDirLen
                                ,   recls_char_t const          *pattern
                                ,   hrecls_progress_fn_t        pfn
                                ,   recls_process_fn_param_t    param
                                ,   recls_uint32_t              flags)
    : m_flags(flags)
    , m_rootDir(calc_rootDir_(cDirParts, rootDir))
    , m_rootDirLen(rootDirLen)
    , m_pfn(pfn)
    , m_param(param)
{
    function_scope_trace("ReclsFileSearch::ReclsFileSearch");

    RECLS_ASSERT(NULL != rootDir);
    RECLS_ASSERT(NULL != pattern);
    RECLS_ASSERT(traits_type::str_len(rootDir) < traits_type::path_max());
    RECLS_ASSERT(NULL != traits_type::str_chr(pattern, traits_type::path_separator()) || traits_type::str_len(pattern) < traits_type::path_max());
    RECLS_ASSERT(m_rootDirLen == traits_type::str_len(rootDir));

#if defined(RECLS_PLATFORM_IS_WIN32)
    RECLS_MESSAGE_ASSERT("Root directory has improper format",  (isalpha(rootDir[0]) && rootDir[1] == ':') || (rootDir[0] == '\\' && rootDir[1] == '\\'));
#elif defined(RECLS_PLATFORM_IS_UNIX)
# if defined(WIN32) && \
     defined(EMULATE_UNIX_ON_WIN32)
    RECLS_MESSAGE_ASSERT("Root directory has improper format",  rootDir[0] == '/' || (isalpha(rootDir[0]) && rootDir[1] == ':') || (rootDir[0] == '\\' && rootDir[1] == '\\'));
# else /* ? really unix */
    RECLS_MESSAGE_ASSERT("Root directory has improper format", rootDir[0] == '/');
    RECLS_MESSAGE_ASSERT("UNIX filesystems do not allow drive specifications", rootDir[1] != ':');
# endif /* ? really unix */
#else /* ? platform*/
# error Platform not recognised
#endif /* platform*/

    // Now start the search
    m_dnode = ReclsFileSearchDirectoryNode::FindAndCreate(m_flags, m_rootDirLen, rootDir, pattern, pfn, param);
}

ReclsFileSearch::~ReclsFileSearch()
{
    function_scope_trace("ReclsFileSearch::~ReclsFileSearch");
}

/* static */ recls_rc_t ReclsFileSearch::Stat(  recls_char_t const  *path
                                            ,   recls_uint32_t      flags
                                            ,   recls_info_t        *phEntry)
{
    function_scope_trace("ReclsFileSearch::Stat");

    return ReclsFileSearchDirectoryNode::Stat(path, flags, phEntry);
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
