/* /////////////////////////////////////////////////////////////////////////////
 * File:        ReclsFtpSearchDirectoryNode.cpp
 *
 * Purpose:     Implementation of the ReclsFtpSearchDirectoryNode class for Win32.
 *
 * Created:     1st June 2004
 * Updated:     3rd October 2006
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2004-2006, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls/recls.h>
#include <recls/assert.h>
#include "recls_util.h"
#include "recls_impl.h"

#if defined(RECLS_DELAY_LOAD_WININET)
# ifdef INETSTL_INCL_H_INETSTL
#  error INETSTL_INCL_H_INETSTL
# endif /* INETSTL_INCL_H_INETSTL */
# include "recls_wininet_dl.h"
#endif /* RECLS_DELAY_LOAD_WININET */

#if defined(RECLS_CHAR_TYPE_IS_WCHAR)
# include <winstl/conversion/char_conversions.hpp>
#endif /* RECLS_CHAR_TYPE_IS_???? */

#include "recls_debug.h"

#include "ReclsSearch.hpp"
#include "ReclsFtpSearchDirectoryNode_win32.hpp"
#include "EntryFunctions.h"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

namespace {

typedef ReclsFtpSearchDirectoryNode::traits_type    traits_type;

stlsoft_ns_using(c_str_ptr)

} // anonymous namespace

/* /////////////////////////////////////////////////////////////////////////////
 * Constants
 */

static const recls_char_t   s_localDir[]                =   RECLS_LITERAL(".");

/* /////////////////////////////////////////////////////////////////////////////
 * Utility functions
 */

inline size_t count_dir_parts(recls_char_t const *begin, recls_char_t const *end)
{
    return count_char_instances(begin, end, traits_type::path_name_separator());
}

/// Creates an entry from the platform-specific record

static recls_info_t create_entryinfo_from_psrecord(recls_char_t const *rootDir, recls_uint32_t flags, traits_type::find_data_type const &findData)
{
    function_scope_trace("create_entryinfo_from_psrecord");

    // size of structure is:
    //
    //    offsetof(struct recls_fileinfo_t, data)
    //  + directory parts
    //  + full path (+ null)
    //  + short name (+ null)

    // Bizarrely, the findData.cFileName can come in either as the file name or as the full path,
    // which means that we're in for some effort to handle that.

    int     isRooted    =   findData.cFileName[0] == '/';
    size_t  cchRootDir  =   traits_type::str_len(rootDir);
    size_t  cchEntry    =   traits_type::str_len(findData.cFileName);
    size_t  cchFile     =   isRooted ? (cchEntry - ((traits_type::str_rchr(findData.cFileName, '/') + 1) - &findData.cFileName[0])) : cchEntry;
    size_t  cchFullPath =   isRooted ? cchEntry : (cchRootDir + cchFile);

    RECLS_ASSERT(cchRootDir > 0);
    RECLS_ASSERT(rootDir[cchRootDir - 1] == traits_type::path_name_separator());

    size_t  cDirParts   =   ((flags & RECLS_F_DIRECTORY_PARTS) == RECLS_F_DIRECTORY_PARTS) ? count_dir_parts(rootDir, rootDir + cchRootDir) : 0;
    size_t  cbPath      =   align_up_size(sizeof(recls_char_t) * (1 + cchFullPath));
    size_t  cbAlt       =   align_up_size(sizeof(recls_char_t) * (1 + RECLS_NUM_ELEMENTS(findData.cAlternateFileName)));
    size_t  cb          =   offsetof(struct recls_fileinfo_t, data)
                        +   cDirParts * sizeof(recls_strptrs_t)
                        +   cbPath
                        +   cbAlt;

    struct recls_fileinfo_t *info   =   const_cast<struct recls_fileinfo_t*>(Entry_Allocate(cb));

    if(NULL != info)
    {
        recls_char_t            *fullPath   =   ::stlsoft::sap_cast<recls_char_t*>(&info->data[cDirParts * sizeof(recls_strptrs_t)]);
        recls_char_t            *altName    =   ::stlsoft::sap_cast<recls_char_t*>(&info->data[cDirParts * sizeof(recls_strptrs_t) + cbPath]);

        // full path
        traits_type::str_copy(fullPath, isRooted ? RECLS_LITERAL("") : rootDir);
        traits_type::str_cat(fullPath, findData.cFileName);
        info->path.begin                    =   fullPath;
        info->path.end                      =   fullPath + cchFullPath;

        debug_printf("entry: %s\n", fullPath);

        // drive, directory, file (name + ext)
        info->drive                         =   '\0';
        info->directory.begin               =   fullPath;
        info->directory.end                 =   fullPath + cchRootDir;
        info->fileName.begin                =   info->directory.end;
#if defined(RECLS_CHAR_TYPE_IS_CHAR)
        info->fileName.end                  =   strrchr(info->directory.end, '.');
#elif defined(RECLS_CHAR_TYPE_IS_WCHAR)
        info->fileName.end                  =   wcsrchr(info->directory.end, L'.');
#else
# error 
#endif /* RECLS_CHAR_TYPE_IS_???? */

        if(NULL != info->fileName.end)
        {
            info->fileExt.begin             =   info->fileName.end + 1;
            info->fileExt.end               =   info->directory.end +  cchFile;
        }
        else
        {
            info->fileName.end              =   info->directory.end +  cchFile;
            info->fileExt.begin             =   info->directory.end +  cchFile;
            info->fileExt.end               =   info->directory.end +  cchFile;
        }

        // determine the directory parts
        recls_char_t const      *p          =   info->directory.begin;
        recls_char_t const      *l          =   info->directory.end;
        struct recls_strptrs_t  *begin      =   ::stlsoft::sap_cast<struct recls_strptrs_t*>(&info->data[0]);

        info->directoryParts.begin          =   begin;
        info->directoryParts.end            =   begin + cDirParts;

        if(info->directoryParts.begin != info->directoryParts.end)
        {
            RECLS_ASSERT((flags & RECLS_F_DIRECTORY_PARTS) == RECLS_F_DIRECTORY_PARTS);

            begin->begin = p;

            for(; p != l; ++p)
            {
                if(*p == traits_type::path_name_separator())
                {
                    begin->end = p + 1;

                    if(++begin != info->directoryParts.end)
                    {
                        begin->begin = p + 1;
                    }
                }
            }
        }

        // alt name
        traits_type::str_copy(altName, findData.cAlternateFileName);
        info->shortFile.begin   =   altName;
        info->shortFile.end     =   altName + traits_type::str_len(altName);

        // attributes
        info->attributes        =   findData.dwFileAttributes;

        // time, size
        info->creationTime      =   findData.ftCreationTime;
        info->modificationTime  =   findData.ftLastWriteTime;
        info->lastAccessTime    =   findData.ftLastAccessTime;
        info->size.u.HighPart   =   findData.nFileSizeHigh;
        info->size.u.LowPart    =   findData.nFileSizeLow;

        // Checks
        RECLS_ASSERT(info->path.begin < info->path.end);
        RECLS_ASSERT(info->directory.begin < info->directory.end);
        RECLS_ASSERT(info->path.begin <= info->directory.begin);
        RECLS_ASSERT(info->directory.end <= info->path.end);
        RECLS_ASSERT(info->fileName.begin <= info->fileName.end);
        RECLS_ASSERT(info->fileExt.begin <= info->fileExt.end);
        RECLS_ASSERT(traits_type::str_len(info->fileExt.begin) == static_cast<size_t>(info->fileExt.end - info->fileExt.begin));
        RECLS_ASSERT(info->fileName.begin < info->fileExt.end);
        RECLS_ASSERT(info->fileName.end <= info->fileExt.begin);
        RECLS_ASSERT(info->directory.end == info->fileName.begin);
    }

    return info;
}

/* /////////////////////////////////////////////////////////////////////////////
 * ReclsFtpSearchDirectoryNode
 */

/* static */ ReclsFtpSearchDirectoryNode::directory_sequence_type::const_iterator ReclsFtpSearchDirectoryNode::select_iter_if_( unsigned long                           b
                                                                                                                            ,   directory_sequence_type::const_iterator trueVal
                                                                                                                            ,   directory_sequence_type::const_iterator falseVal)
{
    // I can't explain it, but Borland does not like the tertiary operator and the copy-ctors of the iterators
    if(b)
    {
        return trueVal;
    }
    else
    {
        return falseVal;
    }
}

/* static */ int ReclsFtpSearchDirectoryNode::essFlags_from_reclsFlags_(recls_uint32_t flags)
{
#if defined(_RECLS_USING_INETSTL_SEARCHSPEC_SEQUENCE)
    // Because Digital Mars 8.40- has a problem, we must access the typedef separately from the enum value
    typedef entry_sequence_type::find_sequence_type sequence_t;
#else /* ? _RECLS_USING_INETSTL_SEARCHSPEC_SEQUENCE */
    typedef entry_sequence_type                     sequence_t;
#endif /* _RECLS_USING_INETSTL_SEARCHSPEC_SEQUENCE */

    int ssFlags = 0;

    if(0 != (flags & RECLS_F_FILES))
    {
        ssFlags |= sequence_t::files;
    }
    if(0 != (flags & RECLS_F_DIRECTORIES))
    {
        ssFlags |= sequence_t::directories;
    }

    return ssFlags;
}

inline /* static */ ReclsFtpSearchDirectoryNode::directory_sequence_type ReclsFtpSearchDirectoryNode::init_directories_(HINTERNET connection, recls_char_t const *rootDir, recls_uint32_t flags)
{
    RECLS_ASSERT(NULL != connection);

    directory_sequence_type     dirs;

    if(flags & RECLS_F_RECURSIVE)
    {
        file_find_sequence_type     directories(connection, rootDir, traits_type::pattern_all(), file_find_sequence_type::directories);

        file_find_sequence_type::const_iterator begin   =   directories.begin();
        file_find_sequence_type::const_iterator end     =   directories.end();

        for(; begin != end; ++begin)
        {
            dirs.push_back((*begin).get_path());
        }
    }

    return dirs;
}

ReclsFtpSearchDirectoryNode::ReclsFtpSearchDirectoryNode(HINTERNET connection, recls_uint32_t flags, recls_char_t const *rootDir, recls_char_t const *pattern)
    : m_connection(connection)
    , m_current(NULL)
    , m_dnode(NULL)
    , m_flags(flags)
    , m_rootDir()
    , m_pattern()
    , m_directories(init_directories_(connection, rootDir, flags))
    , m_directoriesBegin(m_directories.begin())
#if defined(RECLS_USE_NEW_INETSTL_FFS) || \
    defined(_RECLS_SUPPORTS_MULTIPATTERN)
    , m_entries(connection, rootDir, pattern, traits_type::path_separator(), essFlags_from_reclsFlags_(flags))
#else /* ? RECLS_USE_NEW_INETSTL_FFS || _RECLS_SUPPORTS_MULTIPATTERN */
    , m_entries(connection, rootDir, pattern, essFlags_from_reclsFlags_(flags))
#endif /* RECLS_USE_NEW_INETSTL_FFS || _RECLS_SUPPORTS_MULTIPATTERN */
    , m_entriesBegin(m_entries.begin())
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::ReclsFtpSearchDirectoryNode");

    RECLS_ASSERT(NULL != m_connection);

    if(0 == traits_type::str_compare(s_localDir, pattern))
    {
        // This special case handles "." as a search pattern.
        string_type temp(rootDir);

        if( !temp.empty() &&
            temp[temp.length() - 1] == '/')
        {
            temp = string_type(&temp[0], &temp[temp.length() - 1]);
        }

        size_t              cch =   traits_type::get_full_path_name(m_connection, c_str_ptr(temp), 0, NULL);
        string_type         temp2(cch, ' ');
        recls_char_t        *file;

        traits_type::get_full_path_name(m_connection, c_str_ptr(temp), 1 + temp2.length(), &temp2[0], &file);

        m_rootDir.assign(&temp2[0], file);
        m_pattern.assign(file);
    }
    else
    {
        m_rootDir = rootDir;
        if( !m_rootDir.empty() &&
            m_rootDir[m_rootDir.length() - 1] != '/')
        {
            m_rootDir += '/';
        }
        m_pattern = pattern;
    }

    RECLS_ASSERT(STLSOFT_RAW_OFFSETOF(ReclsFtpSearchDirectoryNode, m_entries) < STLSOFT_RAW_OFFSETOF(ReclsFtpSearchDirectoryNode, m_entriesBegin));
    RECLS_ASSERT(STLSOFT_RAW_OFFSETOF(ReclsFtpSearchDirectoryNode, m_directories) < STLSOFT_RAW_OFFSETOF(ReclsFtpSearchDirectoryNode, m_directoriesBegin));

#ifdef STLSOFT_CF_USE_RAW_OFFSETOF_IN_STATIC_ASSERT
    STLSOFT_STATIC_ASSERT(STLSOFT_RAW_OFFSETOF(ReclsFtpSearchDirectoryNode, m_entries) < STLSOFT_RAW_OFFSETOF(ReclsFtpSearchDirectoryNode, m_entriesBegin));
    STLSOFT_STATIC_ASSERT(STLSOFT_RAW_OFFSETOF(ReclsFtpSearchDirectoryNode, m_directories) < STLSOFT_RAW_OFFSETOF(ReclsFtpSearchDirectoryNode, m_directoriesBegin));
#endif /* STLSOFT_CF_USE_RAW_OFFSETOF_IN_STATIC_ASSERT */
}

/* static */ ReclsFtpSearchDirectoryNode *ReclsFtpSearchDirectoryNode::FindAndCreate(   HINTERNET           connection
                                                                                    ,   recls_uint32_t      flags
                                                                                    ,   recls_char_t const  *rootDir
                                                                                    ,   recls_char_t const  *pattern)
{
    ReclsFtpSearchDirectoryNode  *node;

    function_scope_trace("ReclsFtpSearchDirectoryNode::FindAndCreate");

#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
    try
    {
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */
        node = new ReclsFtpSearchDirectoryNode(connection, flags, rootDir, pattern);
#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
    }
    catch(std::bad_alloc &)
    {
        node = NULL;
    }
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */

    if(NULL != node)
    {
        // Ensure that it, or one of its sub-nodes, has matching entries.
        recls_rc_t  rc = node->Initialise();

        if(RECLS_FAILED(rc))
        {
            delete node;

            node = NULL;
        }
    }

    RECLS_ASSERT(NULL == node || node->is_valid());

    return node;
}

ReclsFtpSearchDirectoryNode::~ReclsFtpSearchDirectoryNode()
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::~ReclsFtpSearchDirectoryNode");

    Entry_Release(m_current);

    delete m_dnode;
}

recls_rc_t ReclsFtpSearchDirectoryNode::Initialise()
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::Initialise");

    recls_rc_t  rc;

    RECLS_ASSERT(NULL == m_current);
    RECLS_ASSERT(NULL == m_dnode);

    if(m_entriesBegin != m_entries.end())
    {
        // (i) Try getting a file first,
        m_current = create_entryinfo_from_psrecord(c_str_ptr(m_rootDir), m_flags, (*m_entriesBegin).get_find_data());

        if(NULL == m_current)
        {
            rc = RECLS_RC_OUT_OF_MEMORY;
        }
        else
        {
            rc = RECLS_RC_OK;
        }
    }
    else
    {
        if(m_directoriesBegin == m_directories.end())
        {
            rc = RECLS_RC_NO_MORE_DATA;
        }
        else
        {
            do
            {
                m_dnode = ReclsFtpSearchDirectoryNode::FindAndCreate(m_connection, m_flags, c_str_ptr(*m_directoriesBegin), c_str_ptr(m_pattern));

            } while(NULL == m_dnode && ++m_directoriesBegin != m_directories.end());

            rc = (NULL == m_dnode) ? RECLS_RC_NO_MORE_DATA : RECLS_RC_OK;
        }
    }

    if(RECLS_SUCCEEDED(rc))
    {
        RECLS_ASSERT(is_valid());
    }

    return rc;
}

#ifdef RECLS_ENFORCING_CONTRACTS
recls_bool_t ReclsFtpSearchDirectoryNode::is_valid() const
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::is_valid");

    recls_rc_t  rc  =   RECLS_RC_OK;

#if defined(RECLS_PRAGMA_MESSAGE_SUPPORT) && \
    !defined(RECLS_PRAGMA_MESSAGE_IS_INTRUSIVE)
# pragma message("Flesh these out")
    if(RECLS_SUCCEEDED(rc))
    {

    }
#endif /* compiler */

    // (i) Either we are enumerating files (m_current != NULL) or directories (m_dnode != NULL), but not both
    RECLS_ASSERT(NULL == m_current || NULL == m_dnode);
    // (ii) Either we are enumerating files (m_current != NULL) or there are no more files to enumerate
    RECLS_ASSERT(NULL != m_current || m_entriesBegin == m_entries.end());

    return RECLS_SUCCEEDED(rc);
}
#endif /* RECLS_ENFORCING_CONTRACTS */

recls_rc_t ReclsFtpSearchDirectoryNode::GetNext()
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::GetNext");

    RECLS_ASSERT(is_valid());

    /* Searching operates as follows:
     *
     * 1. Return all the contents of the files
     * 2. Return the contents of the directories.
     *
     * Hence, if m_dnode is non-NULL, we've already searched
     */

    /* States:
     *
     * - Iterating files: m_entriesBegin != m_entries.end(), m_dnode is NULL, m_directoriesBegin != m_directories.end()
     * - Iterating directories: m_directoriesBegin != m_directories.end(), m_dnode is non-NULL, m_current is NULL
     *
     */

    // Invariants

    // (i) Either we are enumerating files (m_current != NULL) or directories (m_dnode != NULL), but not both
    RECLS_ASSERT(NULL == m_current || NULL == m_dnode);
    // (ii) Either we are enumerating files (m_current != NULL) or there are no more files to enumerate
    RECLS_ASSERT(NULL != m_current || m_entriesBegin == m_entries.end());

    recls_rc_t  rc = RECLS_RC_NO_MORE_DATA;

    if(NULL != m_current)
    {
        // Currently enumerating through the files

        RECLS_ASSERT(m_entriesBegin != m_entries.end());
        RECLS_ASSERT(NULL == m_dnode);

        // Advance, and check for end of sequence
        ++m_entriesBegin;

        Entry_Release(m_current);
        if(m_entriesBegin != m_entries.end())
        {
            // Still enumerating, so just update m_current
            m_current = create_entryinfo_from_psrecord(c_str_ptr(m_rootDir), m_flags, (*m_entriesBegin).get_find_data());

            rc = RECLS_RC_OK;
        }
        else
        {
            // No more left in the files sequence, so delete m_current
            m_current = NULL;

            rc = RECLS_RC_NO_MORE_DATA;
        }
    }

    if(NULL == m_current)
    {
        // Now we are either enumerating the directories, or we've already done so

        if(NULL != m_dnode)
        {
            // Currently enumerating the directories
            rc = m_dnode->GetNext();

            if(RECLS_RC_NO_MORE_DATA == rc)
            {
                ++m_directoriesBegin;

                delete m_dnode;

                m_dnode = NULL;
            }
        }

        if(m_directoriesBegin == m_directories.end())
        {
            // Enumeration is complete.
            rc = RECLS_RC_NO_MORE_DATA;
        }
        else
        {
            if(NULL == m_dnode)
            {
                do
                {
                    // Creation of the node will cause it to enter the first enumeration
                    // state. However, if there are no matching

                    RECLS_ASSERT(m_directoriesBegin != m_directories.end());

                    m_dnode = ReclsFtpSearchDirectoryNode::FindAndCreate(m_connection, m_flags, c_str_ptr(*m_directoriesBegin), c_str_ptr(m_pattern));

                    if(NULL != m_dnode)
                    {
                        rc = RECLS_RC_OK;
                    }
                    else
                    {
                        ++m_directoriesBegin;
                    }

                } while(NULL == m_dnode && m_directoriesBegin != m_directories.end());
            }
        }
    }

    RECLS_ASSERT(is_valid());

    return rc;
}

recls_rc_t ReclsFtpSearchDirectoryNode::GetDetails(recls_info_t *pinfo)
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::GetDetails");

    RECLS_ASSERT(is_valid());

    recls_rc_t  rc;

    RECLS_ASSERT(NULL != pinfo);
    RECLS_ASSERT(NULL == m_current || NULL == m_dnode);

    if(NULL != m_current)
    {
        // Currently searching for files from the current directory

        RECLS_ASSERT(NULL == m_dnode);

        rc = Entry_Copy(m_current, pinfo);

#if RECLS_DEBUG_LEVEL >= 2
        {
            size_t      cch =   Recls_GetPathProperty(m_current, NULL, 0);
            string_type buffer(cch, ' ');

            Recls_GetPathProperty(m_current, &buffer[0], 1 + buffer.length());

# if defined(RECLS_CHAR_TYPE_IS_CHAR)
            debug_printf("    [%s]\n", c_str_ptr(buffer));
# elif defined(RECLS_CHAR_TYPE_IS_WCHAR)
            debug_printf("    [%s]\n", static_cast<char const*>(winstl::w2a(buffer)));
# endif /* RECLS_CHAR_TYPE_IS_???? */
        }
#endif /* RECLS_DEBUG_LEVEL >= 2 */
    }
    else if(NULL != m_dnode)
    {
        RECLS_ASSERT(NULL == m_current);

        // Sub-directory searching is active, so get from there.

        rc = m_dnode->GetDetails(pinfo);
    }
    else
    {
        // Enumeration has completed
        rc = RECLS_RC_NO_MORE_DATA;
    }

    RECLS_ASSERT(is_valid());

    return rc;
}

recls_rc_t ReclsFtpSearchDirectoryNode::GetNextDetails(recls_info_t *pinfo)
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::GetNextDetails");

    RECLS_ASSERT(is_valid());
    RECLS_ASSERT(NULL != pinfo);

    recls_rc_t  rc  =   GetNext();

    if(RECLS_SUCCEEDED(rc))
    {
        rc = GetDetails(pinfo);
    }

    RECLS_ASSERT(is_valid());

    return rc;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
